import torch
from typing import Dict, List, Callable, Any

from pylopt.data import OptimiserResult
from pylopt.optimise.optimise_nag import flatten_groups, compute_relative_error

LR_KEY = 'lr'
DEFAULTS_GROUP_ADAM = {LR_KEY: 1e-4, 'betas': (0.9, 0.999), 'weight_decay': 0.0, 'eps': 1e-8}

def create_projected_optimiser(base_optimiser: type[torch.optim.Optimizer]) -> type[torch.optim.Optimizer]:

    # TODO
    #   > take list of projections as input

    class ProjectedOptimiser(base_optimiser):
        def __init__(self, params, *args, **kwargs):
            super().__init__(params, *args, **kwargs)

        def step(self, closure=None):
            loss = super().step(closure)

            with torch.no_grad():
                for group in self.param_groups:
                    for p in group['params']:
                        if not p.requires_grad:
                            continue
                        if hasattr(p, 'orthogonal_projection'):
                            p.data.copy_(p.orthogonal_projection(p))
                        
                        if hasattr(p, 'unit_ball_projection'):
                            p.data.copy_(p.unit_ball_projection(p))

                        if hasattr(p, 'zero_mean_projection'):
                            p.data.copy_(p.zero_mean_projection(p.data))
            return loss

    ProjectedOptimiser.__name__ = 'Projected{:s}'.format(base_optimiser.__name__)
    return ProjectedOptimiser

def harmonise_param_groups_adam(param_groups: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    param_groups_ = []
    for group in param_groups:
        group_ = {'params': [p for p in group['params']],
                  'history': [p.detach().clone().requires_grad_(True) for p in group['params']]}
        for key in group.keys():
            if key != 'params':
                group_[key] = group[key]

        group_[LR_KEY] = group[LR_KEY] if group[LR_KEY] is not None else DEFAULTS_GROUP_ADAM[LR_KEY]
        group_['betas'] = group['betas'] if group['betas'] is not None else DEFAULTS_GROUP_ADAM['betas']
        group_['weight_decay'] = group['weight_decay'] if group['weight_decay'] is not None \
            else DEFAULTS_GROUP_ADAM['weight_decay']
        group_['eps'] = DEFAULTS_GROUP_ADAM['eps']

        param_groups_.append(group_)
    return param_groups_

def step_adam(optimiser: torch.optim.Optimizer, func: Callable, param_groups: List[Dict[str, Any]]) -> torch.Tensor:
    # update history
    for group in param_groups:
        for p, p_old in zip([p_ for p_ in group['params']], [p_ for p_ in group['history']]):
            p_old.data.copy_(p.data.clone())

    # make adam step
    optimiser.zero_grad()
    params_flat = flatten_groups(param_groups)
    with torch.enable_grad():
        loss = func(*params_flat)
    loss.backward()
    optimiser.step()

    return loss

def optimise_adam(func: Callable, param_groups: List[Dict[str, Any]], max_num_iterations: int,
                  rel_tol: float, **unknown_options) -> OptimiserResult:
    num_iterations = max_num_iterations

    param_groups_ = harmonise_param_groups_adam(param_groups)
    optimiser = create_projected_optimiser(torch.optim.SGD)(param_groups_)


    for k in range(0, max_num_iterations):
        _ = step_adam(optimiser, func, param_groups_)

        if rel_tol:
            rel_error = compute_relative_error(param_groups_)
            if rel_error <= rel_tol:
                num_iterations = k + 1
                break

    result = OptimiserResult(solution=param_groups_, num_iterations=num_iterations,
                             loss=func(*flatten_groups(param_groups_)))

    return result





import logsumexpv2 as lse
import math
import torch as th

vmin = -1
vmax = 1
nfeatures = 8
ncomponents = 125
sigma = (vmax - vmin) / (ncomponents - 1)
sigmas = sigma * th.ones((nfeatures, ncomponents),
                         dtype=th.float64,
                         device='cuda')

mus = th.linspace(vmin, vmax, ncomponents, dtype=th.float64, device='cuda')
mus = mus[None].repeat(nfeatures, 1)
mus += th.randn_like(mus) * 1e-2
weights = th.rand((nfeatures, ncomponents), dtype=th.float64, device='cuda') + 1e-5 * th.rand((nfeatures, ncomponents), dtype=th.float64, device='cuda')
sigmas += th.randn_like(sigmas) * 1e-4


def f(x):
    print(x[:, :, None].shape)
    print(mus[None, :, :, None, None].shape)
    d = x[:, :, None] - mus[None, :, :, None, None]
    return -th.logsumexp(
        -(d / sigmas[None, :, :, None, None])**2 / 2 + th.log(
            weights[None, :, :, None, None] / th.sqrt(2 * math.pi * sigmas[None, :, :, None, None]**2)
        ),
        dim=2
    )

b = 16
h, w = 32, 32
x = th.randn(b, nfeatures, h, w, dtype=th.float64, device='cuda', requires_grad=True)
sigmas.requires_grad = True
mus.requires_grad = True
weights.requires_grad = True


f_th = f(x)


from typing import Tuple
import torch
from matplotlib import pyplot as plt
from matplotlib import colormaps as cmaps

def orthogonal_procrustes_projection(x: torch.Tensor, eps: float=1e-7, max_num_iterations: int=5,
                                     rel_tol: float=1e-5) -> torch.Tensor:
    x_flattened = [x[i, 0, :, :].flatten() for i in range(0, x.shape[0])]
    x_stacked = torch.stack(x_flattened, dim=1)
    m, n = x_stacked.shape
    diag = torch.diag(torch.ones(n))

    v_old = torch.zeros_like(x_stacked)
    for k in range(0, max_num_iterations):
        z = torch.matmul(x_stacked, diag)
        q, s, r_h = torch.linalg.svd(z, full_matrices=False)
        v = torch.matmul(q, r_h)

        tmp = torch.matmul(v.transpose(dim0=0, dim1=1), x_stacked)
        diag_elements = torch.diag(tmp)
        diag_elements = torch.clamp(diag_elements, min=eps)
        diag = torch.diag(diag_elements)
        if torch.linalg.norm(v - v_old) < rel_tol:
            break
        v_old = v.clone()

        print(torch.sum((torch.matmul(v, diag) - x_stacked) ** 2))

    x_orthogonal = [torch.unflatten(torch.matmul(v, diag)[:, j], dim=0, sizes=x.shape[-2:]) for j in range(0, v.shape[1])]
    return torch.stack(x_orthogonal, dim=0).unsqueeze(dim=1)

def check_orthogonality(filters):
    pass

def main():
    num_filters = 48
    filter_dim = 7
    filters = torch.rand(num_filters, 1, filter_dim, filter_dim)

    filters_ortho = orthogonal_procrustes_projection(filters)

    # print(torch.linalg.matmul(filters_ortho.transpose(dim0=-2, dim1=-1), filters_ortho))
    #
    # print(torch.linalg.norm(filters_ortho, dim=(-2, -1)))

    fig1, axes = plt.subplots(7, 7, figsize=(11, 11),
                             gridspec_kw={'hspace': 0.9, 'wspace': 0.2})

    for i in range(0, 7):
        for j in range(0, 7):
            filter_idx = i * 7 + j
            if filter_idx < num_filters:
                filter_norm = torch.linalg.norm(filters[filter_idx]).detach().cpu().item()

                # fltr = self._normalise_filter(filters[filter_idx, :, :, :].squeeze().detach().clone())
                axes[i, j].imshow(filters_ortho[filter_idx].squeeze().cpu().numpy(), cmap=cmaps['gray'])

                title = 'idx={:d}, \nnorm={:.3f}'.format(filter_idx, filter_norm)
                axes[i, j].set_title(title, fontsize=8)
                axes[i, j].axis('off')
            else:
                fig1.delaxes(axes[i, j])

    fig2, axes = plt.subplots(7, 7, figsize=(11, 11),
                              gridspec_kw={'hspace': 0.9, 'wspace': 0.2})
    for i in range(0, 7):
        for j in range(0, 7):
            filter_idx = i * 7 + j
            if filter_idx < num_filters:
                filter_norm = torch.linalg.norm(filters[filter_idx]).detach().cpu().item()

                # fltr = self._normalise_filter(filters[filter_idx, :, :, :].squeeze().detach().clone())
                axes[i, j].imshow(filters[filter_idx, 0, :, :].cpu().numpy(), cmap=cmaps['gray'])

                title = 'idx={:d}, \nnorm={:.3f}'.format(filter_idx, filter_norm)
                axes[i, j].set_title(title, fontsize=8)
                axes[i, j].axis('off')
            else:
                fig2.delaxes(axes[i, j])
    plt.show()

if __name__ == '__main__':
    main()



from typing import List, Dict, Any
import torch

def restart_condition_loss_based(param_groups: List[Dict[str, Any]], **kwargs: Dict[str, Any]) -> bool:
    """
    Function implementing loss based restart condition.

    NOTE
    ----
        > Each parameter group must contain the key 'history' of parameter values of the previous iteration.

    :param param_groups: List of dictionaries representing the parameter groups.
    :param kwargs: Keyword arguments - must contain the key-value pair ('func', func), where func is a
        callable representing the loss function.
    :return: Flag indicating if restart condition is satisfied.
    """
    func = kwargs.get('func', None)
    if func is not None:
        f_old = func(*[p for group in param_groups for p in group['history']])
        f_curr = func(*[p for group in param_groups for p in group['params']])

    return f_curr > f_old

def restart_condition_gradient_based(param_groups: List[Dict[str, Any]], **kwargs) -> bool:
    """
    Function implementing gradient based restart condition.

    :param param_groups:
    :param kwargs:
    :return:
    """
    grad_func = kwargs.get('grad_func', None)

    ret_val = False
    if grad_func is not None:
        grads = [g.clone() for g in grad_func(*[p for group in param_groups for p in group['params']])]

        tmp = torch.zeros(1).to(device=grads[-1].device)
        for idx, group in enumerate(param_groups):
            tmp += torch.sum(grads[idx] * (group['params'][-1] - group['history'][-1].to(device=grads[-1].device)))
        ret_val = tmp > 0

    return ret_val

def restart_condition_frequency_based() -> bool:
    pass


import os
import torch
import pathlib
from confuse import Configuration

from pylopt.potential import StudentT

def equal(potential_0: StudentT, potential_1: StudentT) -> bool:
    return (potential_0.num_marginals == potential_1.num_marginals
            and torch.allclose(potential_0.weight_tensor, potential_1.weight_tensor))

def test_save_and_load(tmp_path: pathlib.Path):
    dtype = torch.float64

    file_name = 'potential.pt'

    potential_0 = StudentT().to(dtype=dtype)
    potential_0.save(str(tmp_path), file_name)

    model_path = os.path.join(tmp_path, file_name)
    potential_1 = StudentT.from_file(model_path).to(dtype=dtype)

    assert equal(potential_0, potential_1)

def test_load_from_config():
    dtype = torch.float64

    config = Configuration('unit_test')
    config.set_file(os.path.join('./tests/data', 'test_config.yaml'))

    torch.manual_seed(123)
    potential_0 = StudentT(num_marginals=config['potential']['student_t']['num_marginals'].get(),
                           initialisation_mode=config['potential']['student_t']['initialisation']['mode'].get(),
                           multiplier=config['potential']['student_t']['initialisation']['multiplier'].get(),
                           trainable=config['potential']['student_t']['trainable'].get()).to(dtype=dtype)

    torch.manual_seed(123)
    potential_1 = StudentT.from_config(config).to(dtype=dtype)

    assert equal(potential_0, potential_1)


import torch
from matplotlib import pyplot as plt


from bspline_cuda import quartic_bspline_forward, quartic_bspline_backward
from bspline_cuda.functions import QuarticBSplineFunction


from bilevel_optimisation.utils.config_utils import parse_datatype, load_app_config
from bilevel_optimisation.fields_of_experts import FieldsOfExperts
from bilevel_optimisation.filters import ImageFilter
from bilevel_optimisation.potential import Potential


config = load_app_config('a test only ...', 'example_prediction_III', 'this module')
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
dtype = parse_datatype(config)

image_filter = ImageFilter(config)
potential = Potential.from_config(image_filter.get_num_filters(), config).to(dtype=dtype, device=device)
regulariser = FieldsOfExperts(potential, image_filter)

t = torch.stack([torch.linspace(-5, 5, 111)
                    for _ in range(0, image_filter.filter_dim)]).unsqueeze(dim=1).unsqueeze(dim=0)
t = t.to(dtype=dtype, device=device)

y = potential.forward_negative_log(t, reduce=False)

fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.plot(t[0, 0, 0, :].detach().cpu().numpy(), y[0, 0, 0, :].detach().cpu().numpy())

plt.show()



import torch
from typing import Any, Optional, Dict
from dataclasses import dataclass

@dataclass
class OptimiserResult:
    solution: Any
    num_iterations: int
    loss: torch.Tensor
    meta: Optional[Dict[str, Any]] = None


from pylopt.optimise.optimise_nag import LIP_CONST_KEY
from pylopt.optimise.optimise_nag import optimise_nag, step_nag, optimise_nag_unrolling
from pylopt.optimise.optimise_adam import LR_KEY
from pylopt.optimise.optimise_adam import create_projected_optimiser, optimise_adam, step_adam


from abc import ABC, abstractmethod
import torch
from typing import Dict, Any, Self
from confuse import Configuration
import os

class Potential(ABC, torch.nn.Module):
    """
    Class which is used as base class for FoE-potentials. By design, there is one
    potential function per filter.
    """

    def __init__(self, num_marginals: int):
        """
        Initialisation of an object of class Potential.

        :param num_marginals: Number of marginal potentials required for the FoE-model. By design
            there is one marginal potential per filter.
        """
        super().__init__()
        self.num_marginals = num_marginals

    def get_num_marginals(self) -> int:
        return self.num_marginals

    def state_dict(self, *args, **kwargs) -> Dict[str, Any]:
        state = super().state_dict(*args, **kwargs)
        return state

    @abstractmethod
    def freeze(self) -> None:
        pass

    @abstractmethod
    def unfreeze(self) -> None:
        pass

    @classmethod
    @abstractmethod
    def from_file(cls, path_to_model: str, device: torch.device=torch.device('cpu')):
        pass

    @classmethod
    @abstractmethod
    def from_config(cls, config: Configuration) -> Self:
        pass

    @abstractmethod
    def initialisation_dict(self)  -> Dict[str, Any]:
        pass

    @abstractmethod
    def get_parameters(self) -> torch.Tensor:
        pass

    @abstractmethod
    def forward(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:
        pass

    @staticmethod
    def save(data_dict: Dict[str, Any], path_to_model_dir: str, model_name: str) -> str:
        path_to_model = os.path.join(path_to_model_dir, '{:s}.pt'.format(os.path.splitext(model_name)[0]))
        torch.save(data_dict, path_to_model)
        return path_to_model


from pylopt.regularisers.fields_of_experts.FieldsOfExperts import FieldsOfExperts
from pylopt.regularisers.fields_of_experts.ImageFilter import ImageFilter
from pylopt.regularisers.fields_of_experts.potential.Potential import Potential
from pylopt.regularisers.fields_of_experts.potential.StudentT import StudentT
from pylopt.regularisers.fields_of_experts.potential.Spline import QuarticBSpline


import torch
from abc import abstractmethod

class ProximalOperator(torch.nn.Module):
    def __init__(self, noise_level: float) -> None:
        super().__init__()
        self.noise_level = torch.nn.Parameter(torch.tensor(noise_level), requires_grad=False)

    @abstractmethod
    def forward(self, u: torch.Tensor, tau: torch.Tensor, u_noisy: torch.Tensor) -> torch.Tensor:
        pass

class DenoisingProx(ProximalOperator):
    def __init__(self, noise_level: float) -> None:
        super().__init__(noise_level)

    def forward(self, u: torch.Tensor, tau: torch.Tensor, u_noisy: torch.Tensor) -> torch.Tensor:
        kappa = (tau / self.noise_level ** 2)
        return (kappa * u_noisy + u) / (1 + kappa)


from typing import Any, Tuple, Optional
import numpy as np
import torch
from torch.autograd import Function
from torch.autograd.function import FunctionCtx
from contextlib import nullcontext

from bilevel_optimisation.solver.solve_tridiagonal import solve_tridiagonal
from bilevel_optimisation.utils.Timer import Timer

def spline_forward(x, a, b, c, d):
    return a + x * (b + x * (c + x * d))

def first_order_spline_derivative(x, b, c, d):
    return b + x * (2 * c + 3 * d * x)

def second_order_spline_derivative(x, c, d):
    return 2 * c + 6 * d * x

class SplineBackward(Function):
    @staticmethod
    def bucketise(x: torch.Tensor, box_lower, box_upper, num_nodes):
        x_scaled = (x - box_lower) / (box_upper - box_lower)
        return torch.clamp((x_scaled * (num_nodes - 1)).ceil().long() - 1, min=0, max=num_nodes - 2)

    @staticmethod
    def compute_nodal_grad(grad_out, x, nodal_values, nodes, num_marginals, step_size, box_lower, box_upper, num_nodes):
        index_tensor = SplineBackward.bucketise(x, box_lower, box_upper, num_nodes)
        y = x - nodes[index_tensor]

        bs, f, w, h = x.shape
        idx_flat = index_tensor.view(bs, f, w, h).permute(1, 0, 2, 3).reshape(f, -1)

        nodals = nodal_values.detach().clone().requires_grad_(True)
        with torch.enable_grad():
            nodals, coeffs_1st_order, coeffs_2nd_order, coeffs_3rd_order = NaturalCubicSpline.nodal_values_to_coefficients(nodals, num_nodes, num_marginals, step_size)

            a = nodals.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
            b = coeffs_1st_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
            c = coeffs_2nd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
            d = coeffs_3rd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)

            spline_pot = torch.sum(spline_forward(y, a, b, c, d))
        grad_nodal = torch.autograd.grad(inputs=nodals, outputs=spline_pot)[0]
        return grad_out * grad_nodal, b, c, d, y

    @staticmethod
    def forward(x, nodal_values, nodes, num_marginals, step_size, box_lower, box_upper, num_nodes) -> Tuple[torch.Tensor, ...]:
        index_tensor = SplineBackward.bucketise(x, box_lower, box_upper, num_nodes)
        y = x - nodes[index_tensor]

        bs, f, w, h = x.shape
        idx_flat = index_tensor.view(bs, f, w, h).permute(1, 0, 2, 3).reshape(f, -1)

        # nodals = nodal_values.detach().clone().requires_grad_(True)
        # with torch.enable_grad():
        nodals, coeffs_1st_order, coeffs_2nd_order, coeffs_3rd_order = (
            NaturalCubicSpline.nodal_values_to_coefficients(nodal_values, num_nodes, num_marginals, step_size))

        a = nodals.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        b = coeffs_1st_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        c = coeffs_2nd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        d = coeffs_3rd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)

        #     spline_pot = torch.sum(spline_forward(y, a, b, c, d))
        #     grad_nodal = torch.autograd.grad(inputs=nodals, outputs=spline_pot, create_graph=True)[0]

        return torch.sum(spline_forward(y, a, b, c, d)), a, b, c, d

    @staticmethod
    def setup_context(ctx: Any, inputs: Tuple[Any, ...], output: Tuple[torch.Tensor, ...]) -> Any:
        # x = inputs[0]
        # nodal_values = inputs[1]
        # nodes = inputs[2]
        x, nodal_values, nodes, num_marginals, step_size, box_lower, box_upper, num_nodes = inputs

        ctx.save_for_backward(x, nodal_values, nodes)
        ctx.num_marginals = num_marginals
        ctx.step_size = step_size
        ctx.box_lower = box_lower
        ctx.box_upper = box_upper
        ctx.num_nodes = num_nodes

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        x, nodal_values, nodes = ctx.saved_tensors
        num_marginals = ctx.num_marginals
        step_size = ctx.step_size
        box_lower = ctx.box_lower
        box_upper = ctx.box_upper
        num_nodes = ctx.num_nodes
        return SplineBackwardBackward.apply(grad_outputs[0], x, nodal_values, nodes, num_marginals, step_size, box_lower, box_upper, num_nodes)

class SplineBackwardBackward(Function):
    @staticmethod
    def forward(grad_out, x, nodal_values, nodes, num_marginals, step_size, box_lower, box_upper, num_nodes) -> Any:
        grad_nodal, b, c, d, y = SplineBackward.compute_nodal_grad(grad_out, x, nodal_values, nodes, num_marginals, step_size,
                                                       box_lower, box_upper, num_nodes)

        grad_x = grad_out * first_order_spline_derivative(y, b, c, d)

        return grad_x, grad_out * grad_nodal, *[None] * 6

    @staticmethod
    def setup_context(ctx: Any, inputs: Tuple[Any, ...], output: Any) -> Any:
        grad_out, x, _, a, b, c, d, y, _ = inputs
        ctx.save_for_backward(a, b, c, d, y)
        ctx.grad_out = grad_out

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        a, b, c, d, y = ctx.saved_tensors
        grad_out = ctx.grad_out

        partial_x = grad_outputs[0]
        partial_nodals = grad_outputs[1]

        summand_1 = grad_out * second_order_spline_derivative(y, c, d) * partial_x

        summand_2 = grad_out * ... * partial_nodals
        grad_dx_dx = summand_1 + summand_2

        grad_dx_dnodals

        dx = grad_out * grad_outputs[0] * second_order_spline_derivative(y, c, d)
        dgrad_out = grad_outputs[0] * first_order_spline_derivative(y, b, c, d)

        return dgrad_out, dx, *[None] * 7

# class NaturalSplineAutogradFunction(Function):
#
#     @staticmethod
#     def bucketise(x: torch.Tensor, box_lower, box_upper, num_nodes):
#         x_scaled = (x - box_lower) / (box_upper - box_lower)
#         return torch.clamp((x_scaled * (num_nodes - 1)).ceil().long() - 1, min=0, max=num_nodes - 2)
#
#     @staticmethod
#     def forward(ctx: FunctionCtx, nodes, nodal_values, coeffs_1st_order, coeffs_2nd_order,
#                 coeffs_3rd_order, box_lower, box_upper, num_nodes, x,
#                 grad_outputs=None, order=0) -> Tuple[torch.Tensor, Optional[Any]]:
#         index_tensor = NaturalSplineAutogradFunction.bucketise(x, box_lower, box_upper, num_nodes)
#         y = x - nodes[index_tensor]
#
#         bs, f, w, h = x.shape
#         idx_flat = index_tensor.view(bs, f, w, h).permute(1, 0, 2, 3).reshape(f, -1)
#
#         context = torch.enable_grad() if order == 2 else nullcontext()
#         with context:
#             a = nodal_values.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
#             b = coeffs_1st_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
#             c = coeffs_2nd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
#             d = coeffs_3rd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
#
#             ds_dx = first_order_spline_derivative(y, b, c, d)
#
#         ctx.save_for_backward(y, b, c, d)
#         ctx.order = order
#
#         if order == 0:
#             return a + y * (b + y * (c + y * d)), None
#         elif order == 1:
#             return first_order_spline_derivative(y, b, c, d), None
#         elif order == 2:
#             grad_outputs = torch.ones_like(y) if grad_outputs is None else grad_outputs
#             d2s_mixed = torch.autograd.grad(inputs=nodal_values, outputs=ds_dx, grad_outputs=grad_outputs)
#             return second_order_spline_derivative(y, c, d), d2s_mixed
#         else:
#             raise ValueError
#
#     @staticmethod
#     def backward(ctx: FunctionCtx, *grad_outputs: Any) -> Any:
#         y, b, c, d = ctx.saved_tensors
#         order = ctx.order
#
#         grad = None
#         if order == 0:
#             grad = grad_outputs * first_order_spline_derivative(y, b, c, d)
#         elif order == 1:
#             grad = grad_outputs * second_order_spline_derivative(y, c, d)
#         else:
#             print('problem: nix gut!')
#
#         return *[None] * 8, grad, None, None



class NaturalCubicSplinePotential(torch.nn.Module):

    def __init__(self, num_marginals: int, num_nodes: int, box_lower: float, box_upper: float) -> None:

        super().__init__()
        self.num_marginals = num_marginals

        self.num_nodes = num_nodes
        self.box_lower = box_lower
        self.box_upper = box_upper
        self.register_buffer('nodes', torch.linspace(self.box_lower, self.box_upper, self.num_nodes))
        self.step_size = self.nodes[1] - self.nodes[0]

        vals_1 = torch.tensor([5.0, 2.0, 4.0])
        vals_2 = torch.tensor([-3.0, -2.0, -1.0])
        vals = torch.stack([vals_1, vals_2]).to(device=torch.device('cuda'))

        self.nodal_values = torch.nn.Parameter(vals, requires_grad=True)

    def forward_(self, x):

        index_tensor = SplineBackward.bucketise(x, self.box_lower, self.box_upper, self.num_nodes)

        y = x - self.nodes[index_tensor]

        bs, f, w, h = x.shape
        idx_flat = index_tensor.view(bs, f, w, h).permute(1, 0, 2, 3).reshape(f, -1)

        nodals, coeffs_1st_order, coeffs_2nd_order, coeffs_3rd_order = (
            NaturalCubicSpline.nodal_values_to_coefficients(self.nodal_values, self.num_nodes, self.num_marginals, self.step_size))

        a = nodals.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        b = coeffs_1st_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        c = coeffs_2nd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)
        d = coeffs_3rd_order.gather(1, idx_flat).view(f, bs, w, h).permute(1, 0, 2, 3)


        return torch.sum(spline_forward(y, a, b, c, d))

    def forward(self, x):
        return SplineBackward.apply(x, self.nodal_values, self.nodes, self.num_marginals,
                              self.step_size, self.box_lower, self.box_upper, self.num_nodes)[0]



class NaturalCubicSpline(torch.nn.Module):

    def __init__(self, num_marginals: int, num_nodes: int, box_lower: float, box_upper: float) -> None:
        super().__init__()

        self.num_marginals = num_marginals

        self.num_nodes = num_nodes
        self.box_lower = box_lower
        self.box_upper = box_upper
        self.register_buffer('nodes', torch.linspace(self.box_lower, self.box_upper, self.num_nodes))
        self.step_size = self.nodes[1] - self.nodes[0]

        vals_1 = torch.tensor([5.0, 2.0, 4.0])
        vals_2 = torch.tensor([-3.0, -2.0, -1.0])
        vals = torch.stack([vals_1, vals_2]).to(device=torch.device('cuda'))
        # vals = torch.rand(num_marginals, num_nodes, device=torch.device('cuda'))
        # vals = torch.stack([torch.log(1 + self.nodes ** 2), torch.log(1 + self.nodes ** 2)])

        self.nodal_values = torch.nn.Parameter(vals, requires_grad=True)

        self.diag = torch.ones(self.num_nodes, device=torch.device('cuda'))
        self.diag[1:-1] = 4
        self.diag_super = torch.ones(self.num_nodes - 1, device=torch.device('cuda'))
        self.diag_super[0] = 0
        self.diag_sub = torch.ones(self.num_nodes - 1, device=torch.device('cuda'))
        self.diag_sub[-1] = 0

        self.coeffs_1st_order = torch.zeros(self.num_marginals, self.num_nodes - 1, device=torch.device('cuda'))
        self.coeffs_2nd_order = torch.zeros(self.num_marginals, self.num_nodes - 1, device=torch.device('cuda'))
        self.coeffs_3rd_order = torch.zeros(self.num_marginals, self.num_nodes - 1, device=torch.device('cuda'))

        self._interpolate()

    @staticmethod
    def nodal_values_to_coefficients(nodal_values, num_nodes, num_marginals, step_size):

        diag = torch.ones(num_nodes, device=torch.device('cuda'))
        diag[1:-1] = 4
        diag_super = torch.ones(num_nodes - 1, device=torch.device('cuda'))
        diag_super[0] = 0
        diag_sub = torch.ones(num_nodes - 1, device=torch.device('cuda'))
        diag_sub[-1] = 0

        rhs = torch.zeros(num_marginals, num_nodes, device=torch.device('cuda'))
        rhs[:, 1 : num_nodes - 1] = 3 * (nodal_values[:, 0 : num_nodes - 2] -
                                              2 * nodal_values[:, 1 : num_nodes - 1] +
                                              nodal_values[:, 2::]) / (step_size ** 2)

        coeffs_2nd_order = solve_tridiagonal(diag, diag_super, diag_sub, rhs)
        coeffs_1st_order = ((nodal_values[:, 1 ::] - nodal_values[:, 0 : -1]) / step_size -
                                      step_size * (2 * coeffs_2nd_order[:, 0 : -1] +
                                                        coeffs_2nd_order[:, 1 ::]) / 3)
        coeffs_3rd_order = (coeffs_2nd_order[:, 1 ::] - coeffs_2nd_order[:, 0 : -1]) / (3 * step_size)

        return nodal_values, coeffs_1st_order, coeffs_2nd_order, coeffs_3rd_order

    def _interpolate(self) -> None:
        rhs = torch.zeros(self.num_marginals, self.num_nodes, device=torch.device('cuda'))
        rhs[:, 1 : self.num_nodes - 1] = 3 * (self.nodal_values[:, 0 : self.num_nodes - 2] -
                                              2 * self.nodal_values[:, 1 : self.num_nodes - 1] +
                                              self.nodal_values[:, 2::]) / (self.step_size ** 2)

        self.coeffs_2nd_order = solve_tridiagonal(self.diag, self.diag_super, self.diag_sub, rhs)
        self.coeffs_1st_order = ((self.nodal_values[:, 1 ::] - self.nodal_values[:, 0 : -1]) / self.step_size -
                                      self.step_size * (2 * self.coeffs_2nd_order[:, 0 : -1] +
                                                        self.coeffs_2nd_order[:, 1 ::]) / 3)
        self.coeffs_3rd_order = (self.coeffs_2nd_order[:, 1 ::] - self.coeffs_2nd_order[:, 0 : -1]) / (3 * self.step_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return SplineBackward.apply(x, self.nodal_values, self.nodes, self.num_marginals,
                                    self.step_size, self.box_lower, self.box_upper, self.num_nodes)[0]

        # return NaturalSplineAutogradFunction.apply(self.nodes, self.nodal_values, self.coeffs_1st_order,
        #                                     self.coeffs_2nd_order, self.coeffs_3rd_order,
        #                                     self.box_lower, self.box_upper, self.num_nodes, x)[0]

    def first_derivative(self, x) -> torch.Tensor:
        # return NaturalSplineAutogradFunction.apply(self.nodes, self.nodal_values, self.coeffs_1st_order,
        #                                     self.coeffs_2nd_order, self.coeffs_3rd_order,
        #                                     self.box_lower, self.box_upper, self.num_nodes, x, None, 1)[0]
        pass


    def second_derivative(self, x) -> torch.Tensor:
        # return NaturalSplineAutogradFunction.apply(self.nodes, self.nodal_values, self.coeffs_1st_order,
        #                                            self.coeffs_2nd_order, self.coeffs_3rd_order,
        #                                            self.box_lower, self.box_upper, self.num_nodes, x, None, 2)
        pass


def example_tridiag():
    d = 4.0 * torch.ones(4).to(dtype=torch.float32)
    d_super = torch.ones(3).to(dtype=torch.float32)
    d_sub = torch.ones(3).to(dtype=torch.float32)

    rhs_1 = torch.arange(0, 4).to(dtype=torch.float32)
    rhs_2 = torch.arange(4, 8).to(dtype=torch.float32)
    rhs = torch.stack([rhs_1, rhs_2]).to(dtype=torch.float32)

    x = solve_tridiagonal(d, d_super, d_sub, rhs)

    A = torch.diag(d) + torch.diag(d_super, diagonal=1) + torch.diag(d_sub, diagonal=-1)

    x_1 = torch.linalg.solve(A, rhs_1)
    x_2 = torch.linalg.solve(A, rhs_2)

    print(x)

    print(x_1)
    print(x_2)

def example_spline():
    spline = NaturalCubicSpline(2, 3, 5, 9)
    spline.to(device=torch.device('cuda'))

    t0 = torch.linspace(5, 9, 111)
    t1 = torch.linspace(5, 9, 111)
    t = torch.stack([t0, t1]).unsqueeze(dim=1).unsqueeze(dim=0).to(device=torch.device('cuda')).requires_grad_(True)

    y = spline(t)
    dy_dt = torch.autograd.grad(inputs=t, outputs=torch.sum(y), create_graph=True)[0]
    d2y_dt2 = torch.autograd.grad(inputs=t, outputs=dy_dt, grad_outputs=torch.ones_like(t))[0]

        # x = torch.ones_like(t).requires_grad_(True)
        # def func(z):
        #     return SplineBackward.apply(x, spline.nodes, spline.nodal_values, spline.coeffs_1st_order,
        #                                 spline.coeffs_2nd_order, spline.coeffs_3rd_order, spline.box_lower,
        #                                 spline.box_upper, spline.num_nodes)[0]
        # torch.autograd.gradcheck(func, (x,))


        # dy_dt = spline.first_derivative(t)
        # d2y_d2t, d2y_mixed = spline.second_derivative(t)


    from matplotlib import pyplot as plt
    fig = plt.figure()
    ax_1 = fig.add_subplot(1, 3, 1)
    ax_1.plot(t0.numpy(), y[0, 0, 0, :].squeeze().cpu().detach().numpy())

    t00 = torch.linspace(5, 7, 55)
    p_0 = 5 - 2.125 * (t00 - 5) + 0.1562 * (t00 - 5) ** 3

    t11 = torch.linspace(7, 9, 55)
    p_1 = 2 - 0.25 * (t11 - 7) + 0.9375 * (t11 - 7) ** 2 - 0.1562 * (t11 - 7) ** 3

    ax_1.plot(t00.numpy(), p_0.squeeze().cpu().detach().numpy(), color='magenta')
    ax_1.plot(t11.numpy(), p_1.squeeze().cpu().detach().numpy(), color='cyan')

    ax_1.plot(t1.numpy(), y[0, 1, 0, :].squeeze().cpu().detach().numpy())



    ax_2 = fig.add_subplot(1, 3, 2)
    ax_2.plot(t0.numpy(), dy_dt[0, 0, 0, :].squeeze().cpu().detach().numpy())
    ax_2.plot(t1.numpy(), dy_dt[0, 1, 0, :].squeeze().cpu().detach().numpy())

    dp_0 =  - 2.125 + 3 * 0.1562 * (t00 - 5) ** 2
    dp_1 = - 0.25 + 2 * 0.9375 * (t11 - 7) - 3 * 0.1562 * (t11 - 7) ** 2

    ax_2.plot(t00.numpy(), dp_0.squeeze().cpu().detach().numpy(), color='magenta')
    ax_2.plot(t11.numpy(), dp_1.squeeze().cpu().detach().numpy(), color='cyan')

    d2p_0 = 6 * 0.1562 * (t00 - 5)
    d2p_1 = 2 * 0.9375 - 6 * 0.1562 * (t11 - 7)

    ax_3 = fig.add_subplot(1, 3, 3)
    ax_3.plot(t0.numpy(), d2y_dt2[0, 0, 0, :].squeeze().cpu().detach().numpy())
    ax_3.plot(t1.numpy(), d2y_dt2[0, 1, 0, :].squeeze().cpu().detach().numpy())
    ax_3.plot(t00.numpy(), d2p_0.squeeze().cpu().detach().numpy(), color='magenta')
    ax_3.plot(t11.numpy(), d2p_1.squeeze().cpu().detach().numpy(), color='cyan')

    plt.show()

def evaluate_spline(num_predictions, device, gradients):

    spline = NaturalCubicSpline(48, 33, 5, 9)
    spline.to(device=torch.device('cuda'))
    spline._interpolate()

    x = torch.rand(10, 48, 256, 256, device=device, requires_grad=gradients)

    spline = torch.compile(spline, mode='max-autotune', backend='inductor', dynamic=True,
								 fullgraph=True)
    #
    #
    timings = []
    for _ in range(0, num_predictions):
        with Timer(device) as t:
            with torch.no_grad():
                y = spline(x)
        timings.append(t.time_delta())

    print('median: {:.5f}'.format(np.median(timings)))
    print('mean: {:.5f}'.format(np.mean(timings)))
    print('mean (tail): {:.5f}'.format(np.mean(timings[5::])))

    print('# ### ##############################################################')

    with torch.no_grad():
        spline.nodal_values.copy_(torch.ones(48, 33, device=torch.device('cuda')))
    spline._interpolate()

    timings = []
    for _ in range(0, 10):
        with Timer(device) as t:
            y = torch.sum(spline(x))
            dy_dx = torch.autograd.grad(inputs=x, outputs=y)
        timings.append(t.time_delta())

    print('median: {:.5f}'.format(np.median(timings)))
    print('mean: {:.5f}'.format(np.mean(timings)))
    print('mean (tail): {:.5f}'.format(np.mean(timings[5::])))

def second_derivative():
    spline = NaturalCubicSpline(2, 3, 5, 9)

    spline._interpolate()
    spline.to(device=torch.device('cuda'))


    t0 = torch.linspace(5, 9, 111)
    t1 = torch.linspace(5, 9, 111)
    t = torch.stack([t0, t1]).unsqueeze(dim=1).unsqueeze(dim=0).to(device=torch.device('cuda')).requires_grad_(True)

    print('# ### ############################################################################')
    print('first derivative')
    with torch.enable_grad():
        y = torch.sum(spline(t))
    dy_dx = torch.autograd.grad(inputs=t, outputs=y)[0]
    dy_dx_ = spline.first_derivative(t)

    print(dy_dx.shape)
    print(torch.linalg.norm(dy_dx - dy_dx_))

    print('# ### ############################################################################')

    dz_dt_ = spline.second_derivative(t)

    with torch.enable_grad():
        z = torch.sum(spline.first_derivative(t))
    dz_dt = torch.autograd.grad(inputs=t, outputs=z)[0]

    print(dz_dt.shape)
    print(torch.linalg.norm(dz_dt - dz_dt_))

    print('# ### ############################################################################')
    with torch.enable_grad():
        y = torch.sum(spline(t))
    dy_dx = torch.autograd.grad(inputs=spline.parameters(), outputs=y)[0]



    # with torch.enable_grad():
    #     y = torch.sum(spline(t))
    #     dy_dx = torch.autograd.grad(inputs=t, outputs=y, create_graph=True)
    # d2y_d2x = torch.autograd.grad(inputs=t, outputs=dy_dx, create_graph=True, grad_outputs=torch.rand_like(t))

class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()



def spline_pot_example():
    pot = NaturalCubicSplinePotential(2, 3, 5, 9)
    pot.to(device=torch.device('cuda'))
    #
    bs = 10
    f = 48
    w = 256
    h = 256
    x = torch.rand(bs, f, w, h).cuda()
    x.requires_grad_(True)
    #
    # y = pot.forward_(x)
    # grad_x = torch.autograd.grad(inputs=x, outputs=y, retain_graph=True, create_graph=True)[0]
    #
    # z = torch.sum(grad_x)
    # dz_dnodal = torch.autograd.grad(inputs=[p for p in pot.parameters() if p.requires_grad], outputs=z)
    #
    # asd

    num_trials = 100

    times_1 = []
    for i in range(0, num_trials):
        with Timer(torch.device('cuda')) as t:
            y = pot(x)
            #grad_x = torch.autograd.grad(inputs=x, outputs=y, retain_graph=True)[0]
        times_1.append(t.time_delta())

    grad_nodals = torch.autograd.grad(inputs=pot.nodal_values, outputs=y)[0]

    x_ = x.detach().clone().requires_grad_(True)
    times_2 = []
    for i in range(0, num_trials):
        with Timer(torch.device('cuda')) as t:
            y_ = pot.forward_(x_)
            #grad_x_ = torch.autograd.grad(inputs=x_, outputs=y_, retain_graph=True)[0]
        times_2.append(t.time_delta())

    print(np.mean(times_1))
    print(np.mean(times_2))

    print(torch.linalg.norm(y - y_))
    # print(torch.linalg.norm(grad_x_ - grad_x))


    print('')


if __name__ == '__main__':
    # example_tridiag()
    example_spline()

    # spline_pot_example()

    evaluate_spline(100, torch.device('cuda'), True)

    # second_derivative()






import torch
from typing import List, Dict, Any, Callable, Optional, Tuple
import math

from pylopt.data import OptimiserResult

MAX_NUM_ITERATIONS_DEFAULT = 1000
LIP_CONST_KEY = 'lip_const'
DEFAULTS_GROUP = {'alpha': 1e-4, 'beta': 0.71, 'theta': 0.0, LIP_CONST_KEY: 1e5, 'max_num_backtracking_iterations': 10}

def harmonise_param_groups_nag(param_groups: List[Dict[str, Any]], break_graph: bool=False) -> List[Dict[str, Any]]:
    param_groups_merged = []
    for group in param_groups:
        group_ = {'history': [p.detach().clone() for p in group['params']],
                  'name': group.get('name', '')}
        if break_graph:
            group_['params'] = []
            for p in group['params']:
                p_ = p.detach().clone().requires_grad_(True)
                if hasattr(p, 'prox'):
                    setattr(p_, 'prox', p.prox)
                if hasattr(p, 'proj'):
                    setattr(p_, 'proj', p.proj)
                group_['params'].append(p_)
        else:
            group_['params'] = [p for p in group['params']]

        group_['alpha'] = group['alpha']
        group_['beta'] = group['beta']
        if all(item is not None for item in group_['beta']):
            pass
        elif all(item is None for item in group_['beta']):
            group_['theta'] = DEFAULTS_GROUP['theta']
        else:
            group_['beta'] = [item if item is not None else DEFAULTS_GROUP['beta'] for item in group['beta']]

        if all(item is not None for item in group_['alpha']):
            pass
        elif all(item is None for item in group_['alpha']):
            group_[LIP_CONST_KEY] = [item if item is not None else DEFAULTS_GROUP[LIP_CONST_KEY]
                                   for item in group[LIP_CONST_KEY]]
        else:
            group_['alpha'] = [item if item is not None else DEFAULTS_GROUP['alpha'] for item in group['alpha']]

        group_['max_num_backtracking_iterations'] = group.get('max_num_backtracking_iterations',
                                                              DEFAULTS_GROUP['max_num_backtracking_iterations'])
        param_groups_merged.append(group_)
    return param_groups_merged

def flatten_groups(param_groups: List[Dict[str, Any]]) -> List[torch.Tensor]:
    return [p for group in param_groups for p in group['params']]

def compute_relative_error(param_groups: List[Dict[str, Any]], eps: float=1e-7) -> torch.Tensor:
    error = 0.0
    n = 0
    for group in param_groups:
        for p, p_old in zip(group['params'], group['history']):
            n += p.shape[0]
            error += torch.sum(torch.sqrt(torch.sum((p - p_old) ** 2, dim=(-2, -1)))
                               / torch.sqrt(torch.sum(p_old ** 2, dim=(-2, -1)) + eps))
    return error / n

def make_intermediate_step(group: Dict[str, Any], in_place: bool=True) -> Dict[str, Any]:
    """
    Function which computes the intermediate step of the Nesterov scheme.

    :param group: Parameter group to be updated
    :param in_place: Flag indicating if update shall be updated in place. For the unrolling scheme in place
        updates cannot be used, since computational graph would break.
    :return: Group of updated parameters.
    """
    param = group.get('params')[0]
    param_old = group.get('history')[0]

    if all(item is not None for item in group['beta']):
        beta = torch.tensor(group['beta'], dtype=param.dtype, device=param.device)
    else:
        theta = group['theta']
        theta_new = 0.5 * (1 + math.sqrt(1 + 4 * (theta ** 2)))
        beta = torch.tensor((theta - 1) / theta_new, dtype=param.dtype, device=param.device)
        group['theta'] = theta_new

    # make intermediate step
    momentum = beta.reshape(-1, *[1] * (param.dim() - 1)) * (param - param_old)
    if in_place:
        param_old.copy_(param.data)
        param.add_(momentum)
        return group
    else:
        param_old = param.detach().clone()
        param_new = param + momentum
        group_new = {**group, 'params': [param_new], 'history': [param_old]}
        return group_new

def make_gradient_step(param: torch.nn.Parameter, grads: torch.Tensor, alpha: torch.Tensor,
                       in_place: bool=True) -> torch.Tensor:
    """
    Function which performs a gradient step on param.

    NOTE
    ----
        > In case no in place update is made, projections will not be applied. This is because the non in-place update
            is required only for the unrolling scheme of the lower level problem, where no projections are needed.

    TODO
    ----
        > Handle application of projections and prox-operator better.

    :param param: PyTorch parameter to be updated.
    :param grads: Gradient of objective function w.r.t. param
    :param alpha: Step size
    :param in_place: Flag indicating if update is performed in place. For updates within the unrolling scheme,
        updates can't be made in place.
    :return:
    """
    alpha_ = alpha.reshape(-1, *[1] * (param.dim() - 1))
    if in_place:
        param.sub_(alpha_ * grads)
        if hasattr(param, 'prox'):
            param.data.copy_(param.prox(param.data, alpha_))
        if hasattr(param, 'orthogonal_projection'):
            param.data.copy_(param.orthogonal_projection(param))
        if hasattr(param, 'unit_ball_projection'):
            param.data.copy_(param.unit_ball_projection(param))
        if hasattr(param, 'zero_mean_projection'):
            param.data.copy_(param.zero_mean_projection(param))
    else:
        param = param - alpha_ * grads
        if hasattr(param, 'prox'):
            param = param.prox(param.data, alpha_)

    return param

def backtracking_line_search(param: torch.nn.Parameter, grads: torch.Tensor, closure: Callable,
                             lip_const: torch.Tensor, rho_1: float, rho_2: float,
                             max_num_iterations: int=10) -> torch.Tensor:
    """
    Function which performs backtracking line search for the non-unrolling update to find a step size giving
    sufficient descent.

    :param param: Parameter to be updated.
    :param grads: Gradient of objective function w.r.t. param
    :param closure: Closure function, which evaluates the loss function at the given parameters.
    :param lip_const: Current set of Lipschitz constants.
    :param rho_1: Decrease factor for Lipschitz constant
    :param rho_2: Increase factor for Lipschitz constant
    :param max_num_iterations: Maximal number of backtracking iterations.
    :return: Updated set of Lipschitz constants.
    """
    param_orig = param.data.clone()
    loss = closure()

    for k in range(0, max(1, max_num_iterations)):
        _ = make_gradient_step(param, grads, 1 / lip_const)

        loss_new = closure()
        quadr_approx = (loss + torch.sum(grads * (param - param_orig), dim=tuple(range(1, grads.ndim)) or None)
                        + 0.5 * lip_const * torch.sum((param - param_orig) ** 2, dim=tuple(range(1, grads.ndim)) or None))
        sufficient_descent_met = loss_new <= quadr_approx

        param.data.copy_(param_orig)
        if sufficient_descent_met.all():
            break
        else:
            lip_const = torch.where(sufficient_descent_met, lip_const, lip_const * rho_2)

    _ = make_gradient_step(param, grads, 1 / lip_const)
    return torch.where(sufficient_descent_met, lip_const * rho_1, lip_const)

def step_nag(func: Callable, grad_func: Callable, param_groups: List[Dict[str, Any]],
                   rho_1: float=0.9, rho_2: float=2.0) -> torch.Tensor:
    """
    Function which performs a single NAG or NAPG step.

    NOTE
    ----
    > Implementation assumes that parameters are given in terms of parameter groups, i.e. a list
        of dictionaries, where each dictionary represents a parameter group. Every parameter group
        must have the following key, value pairs:
            - 'params': List of PyTorch parameters
            - 'alpha': List of floats, or list of Nones indicating the line search strategy w.r.t. to
                batch-optimisation or per-sample optimisation. If a list of floats is specified, these values are
                used as constant step sizes; otherwise backtracking line search is applied.
            - 'beta': List of floats, or list of Nones representing the momentum parameters per optimisation
                variable (batch or samples). If it contains a None, Nesterov's momentum schedule
                is applied for all optimisation variables.
            - 'lip_const': List of floats, or list of Nones specifying the Lipschitz constants used in the
                backtracking line search scheme. If one the provided value is None, the default value is used.
        Projection operators, and the proximal operator are assumed to be given on parameter level in terms
        of one of the attributes 'zero_mean_projection', 'orthogonal_projection', 'prox'.
    > Optimisation procedure is alternating w.r.t. parameter groups. This means, that
        for each group, sequentially, gradients are computed and applied.
    > This implementation is feasible for solving the lower level and the upper level problem.

    :param func: Function representing the objective function
    :param grad_func: Function representing the gradient of func
    :param param_groups: Groups of parameters to be optimised.
    :param rho_1: Decrease factor for Lipschitz constant
    :param rho_2: Increase factor for Lipschitz constant
    :return: Loss at new iterate
    """
    for idx, group in enumerate(param_groups):
        _ = make_intermediate_step(group)

        params_flat = flatten_groups(param_groups)
        params_flat_ = [p.requires_grad_(True) for p in params_flat]

        # NOTE
        #   > Cloning the return value of grad_func is necessary to prevent
        #       the following error when applying torch.compile():
        #
        #           [...]
        #           RuntimeError: Error: accessing tensor output of CUDAGraphs that has been overwritten
        #           by a subsequent run.
        #           [...]
        #           To prevent overwriting, clone the tensor outside of torch.compile() or call
        #           torch.compiler.cudagraph_mark_step_begin() before each model invocation.
        #
        grads = grad_func(*params_flat_)[idx].clone()

        param = group.get('params')[0]
        if all(item is not None for item in group['alpha']):
            # use constant step size(s)
            alpha = torch.tensor(group['alpha'], dtype=param.dtype, device=param.device)
            _ = make_gradient_step(param, grads, alpha)
        else:
            # apply backtracking line search
            def closure():
                return func(*flatten_groups(param_groups))

            lip_const = torch.tensor(group[LIP_CONST_KEY], dtype=param.dtype, device=param.device)
            lip_const = backtracking_line_search(param, grads, closure, lip_const, rho_1, rho_2,
                                                 group['max_num_backtracking_iterations'])
            group[LIP_CONST_KEY] = lip_const.tolist()

    return torch.sum(func(*flatten_groups(param_groups)))

@torch.no_grad()
def optimise_nag(func: Callable, grad_func: Callable, param_groups: List[Dict[str, Any]],
                 max_num_iterations: int=MAX_NUM_ITERATIONS_DEFAULT,
                 rel_tol: Optional[float]=None, **unknown_options) -> OptimiserResult:
    """
    Implementation of Nesterov's accelerated gradient method.

    :param func: Callable representing the loss function
    :param grad_func: Callable representing the gradient of func
    :param param_groups: List of dictionaries, where each dictionary represents a parameter group. See step_nag_lower().
    :param max_num_iterations: Maximal number of iterations to be performed.
    :param rel_tol: Tolerance, which if provided, is used as early stopping criterion.
    :param unknown_options:
    :return: Instance of class OptimiserResult containing the computed the solution, the final loss, the
        number of iterations which were performed and a status message.
    """
    num_iterations = max_num_iterations
    param_groups_ = harmonise_param_groups_nag(param_groups, break_graph=True)

    loss = -1 * torch.ones(1)
    for k in range(0, max_num_iterations):

        loss = step_nag(func, grad_func, param_groups_)

        if rel_tol:
            rel_error = compute_relative_error(param_groups_)

            if rel_error <= rel_tol:
                num_iterations = k + 1
                break

    result = OptimiserResult(solution=param_groups_, num_iterations=num_iterations,
                             loss=loss)
    return result

@torch.no_grad()
def backtracking_line_search_unrolling(param_groups: List[Dict[str, Any]], group_idx: int, grads: torch.Tensor,
                                       closure: Callable, rho_1: float=0.9, rho_2: float=2.0,
                                       max_num_iterations: int=10) -> Tuple[torch.Tensor, torch.Tensor]:
    group = param_groups[group_idx]
    param = group.get('params')[0].clone()
    param_orig = group['params'][0].clone()

    loss = closure(param_groups)

    lip_const = torch.tensor(group[LIP_CONST_KEY], dtype=param.dtype, device=param.device)
    for k in range(0, max(1, max_num_iterations)):
        param_new = make_gradient_step(param, grads, 1 / lip_const, in_place=False)

        group_new = {**group, 'params': [param_new]}
        param_groups[group_idx] = group_new
        loss_new = closure(param_groups)

        quadr_approx = (loss + torch.sum(grads * (param_new - param_orig))
                        + 0.5 * lip_const * torch.sum((param_new - param_orig) ** 2))
        sufficient_descent_met = loss_new <= quadr_approx

        group_orig = {**group, 'params': [param_orig]}
        param_groups[group_idx] = group_orig
        if sufficient_descent_met.all():
            break
        else:
            lip_const = torch.where(sufficient_descent_met, lip_const, lip_const * rho_2)

    lip_const_new = torch.where(sufficient_descent_met, lip_const * rho_1, lip_const)

    return lip_const, lip_const_new

def optimise_nag_unrolling(func: Callable, grad_func: Callable, param_groups: List[Dict[str, Any]],
                           max_num_iterations: int=30, rel_tol: Optional[float]=None,
                           **unknown_options) -> OptimiserResult:
    """
    Implementation of the unrolling scheme for solving the lower problem using NAG or NAPG.

    :param func: Callable representing the objective function
    :param grad_func: Callable representing the gradient function of grad
    :param param_groups: List of dictionaries, where each dictionary represents a parameter group. See step_nag_lower().
    :param max_num_iterations: Maximal number of iterations to be performed. Per default only 30 iterations are
        performed.
    :param rel_tol: Tolerance, which if provided, is used for early stopping.
    :param unknown_options:
    :return: Instance of class OptimiserResult containing the computed the solution, the final loss, the
        number of iterations which were performed and a status message.
    """
    num_iterations = max_num_iterations
    param_groups_ = harmonise_param_groups_nag(param_groups, break_graph=False)

    param_groups_current = param_groups_
    for k in range(0, max_num_iterations):
        param_groups_new = []
        for idx, group in enumerate(param_groups_current):
            param_groups_current[idx] = make_intermediate_step(group, in_place=False)

            params_flat = flatten_groups(param_groups_current)
            params_flat_ = [p.requires_grad_(True) for p in params_flat]
            grads = grad_func(*params_flat_)[idx]

            param = group.get('params')[0]
            if all(item is not None for item in group['alpha']):
                alpha = torch.tensor(group['alpha'], dtype=param.dtype, device=param.device)
                param_new = make_gradient_step(param, grads, alpha, in_place=False)
                group_new = {**group, 'params': [param_new]}
            else:
                def closure(_param_groups: List[Dict[str, Any]]) -> torch.Tensor:
                    return func(*flatten_groups(_param_groups))

                lip_const, lip_const_new = backtracking_line_search_unrolling(param_groups_current, idx, grads,
                                                                              closure,
                                                                              max_num_iterations=
                                                                              group['max_num_backtracking_iterations'])
                param_new = make_gradient_step(param, grads, 1 / lip_const, in_place=False)
                group_new = {**group, 'params': [param_new], LIP_CONST_KEY: lip_const_new.tolist()}

            param_groups_new.append(group_new)
        param_groups_current = param_groups_new

        if rel_tol:
            rel_error = compute_relative_error(param_groups_current)

            if rel_error <= rel_tol:
                num_iterations = k + 1
                break

    return OptimiserResult(solution=param_groups_current, num_iterations=num_iterations,
                             loss=func(*flatten_groups(param_groups_current)))




import torch
import time

class Timer:
    """
    Class which is used to measure elapsed on CPU or GPU. Time is measured in unit [ms].
    """
    def __init__(self, device: torch.device) -> None:
        self._device = device
        self._delta_time = -1.0

    def __enter__(self):
        if self._device.type == 'cuda':
            self._start = torch.cuda.Event(enable_timing=True)
            self._end = torch.cuda.Event(enable_timing=True)
            torch.cuda.synchronize()
            self._start.record()
        else:
            self._time_start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._device.type == 'cuda':
            self._end.record()
            torch.cuda.synchronize()
            self._delta_time = self._start.elapsed_time(self._end)
        else:
            self._delta_time = time.time() - self._time_start
            self._delta_time *= 1000

    def time_delta(self) -> float:
        return self._delta_time



from typing import Optional, Dict, Any, Callable, Tuple, List
from abc import ABC, abstractmethod
import torch
from torch.utils.tensorboard import SummaryWriter
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
import os
import math
from matplotlib import pyplot as plt
from matplotlib import colormaps as cmaps
import io
from PIL import Image
import logging
import numpy as np
from confuse import Configuration
import pandas as pd

from pylopt.bilevel_problem.parameter_groups import PARAM_GROUP_NAME_KEY
from pylopt.dataset.dataset_utils import collate_function
from pylopt.dataset.ImageDataset import TestImageDataset
from pylopt.energy import Energy, MeasurementModel
from pylopt.lower_problem import solve_lower
from pylopt.optimise import LIP_CONST_KEY, LR_KEY
from pylopt.regularisers.fields_of_experts.FieldsOfExperts import FieldsOfExperts
from pylopt.utils.evaluation_utils import compute_psnr
from pylopt.utils.Timer import Timer

def figure_to_tensor(fig: plt.Figure) -> torch.Tensor:
    buf = io.BytesIO()
    fig.savefig(buf, format='png', bbox_inches='tight')
    buf.seek(0)
    image = Image.open(buf).convert('RGB')
    return transforms.ToTensor()(image)

def compute_moving_average(data: np.ndarray, window: int) -> np.ndarray:
    lst = [np.nan for _ in range(0, window)]
    for i in range(0, len(data) - window + 1):
        lst.append(np.mean(data[i : i + window]))
    return np.array(lst)

class Callback(ABC):
    """
    Base callback class.
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def on_step(self, step: int, regulariser: Optional[FieldsOfExperts]=None,
                loss: Optional[torch.Tensor]=None, **kwargs) -> None:
        pass

    @abstractmethod
    def on_train_begin(self, regulariser: Optional[FieldsOfExperts]=None, **kwargs) -> None:
        pass

    @abstractmethod
    def on_train_end(self) -> None:
        pass

class SaveModel(Callback):
    """
    Callback which saves current model (image filter and potential) to disk on every
    training/learning step.
    """
    def __init__(self, path_to_data_dir: str, save_freq: int = 2) -> None:
        super().__init__()

        self.path_to_model_dir = os.path.join(path_to_data_dir, 'models')
        self.save_freq = save_freq

    def on_step(self, step: int, regulariser: Optional[FieldsOfExperts]=None,
                loss: Optional[torch.Tensor]=None, **kwargs) -> None:
        if step % self.save_freq == 0 and regulariser is not None:
            if not os.path.exists(self.path_to_model_dir):
                os.makedirs(self.path_to_model_dir, exist_ok=True)
            regulariser.get_image_filter().save(self.path_to_model_dir, 'filters_iter_{:d}.pt'.format(step))
            regulariser.get_potential().save(self.path_to_model_dir, 'potential_iter_{:d}.pt'.format(step))

    def on_train_begin(self, regulariser: Optional[FieldsOfExperts]=None, **kwargs) -> None:
        pass

    def on_train_end(self) -> None:
        pass

class PlotFiltersAndPotentials(Callback):
    """
    Callback which creates plots of current filters and potentials. The plots are saved to disk and
    written to tensorboard.

    NOTE
    ----
        > The j-th potential is plotted on the range [q_low, q_hiqh] of the j-th filter response, where
            q_low, q_high denote the 2.5% and 97.5% quantile respectively of the response of the j-th filter applied to a
            natural image.
    """
    def __init__(self, dataset: TestImageDataset, path_to_data_dir: str, plotting_freq: int = 2,
                 tb_writer: Optional[SummaryWriter]=None, p_low: float=0.025, p_high: float=0.975) -> None:
        """
        Initialisation of filter and potential plotting callback.

        :param dataset: Test image dataset containing natural images which is used to determine the quantile
            range on which the potentials are plotted.
        :param path_to_data_dir: Path to directory where data is stored.
        :param plotting_freq: Frequency on which plots are generated.
        :param tb_writer: Tensorboard writer.
        :param p_low: Probability corresponding to the quantile q_low
        :param p_high: Probability corresponding to the quantile q_high
        """
        super().__init__()
        self.test_loader = DataLoader(dataset, batch_size=len(dataset), shuffle=False,
                                      collate_fn=lambda x: collate_function(x, crop_size=-1))
        self.p_low = p_low
        self.p_high = p_high

        self.path_to_filter_plot_dir = os.path.join(path_to_data_dir, 'filters')
        self.path_to_potential_plot_dir = os.path.join(path_to_data_dir, 'potentials')

        self.tb_writer = tb_writer

        for pth in [self.path_to_filter_plot_dir, self.path_to_potential_plot_dir]:
            if not os.path.exists(pth):
                os.makedirs(pth, exist_ok=True)

        self.plotting_freq = plotting_freq

    def on_step(self, step: int, regulariser: Optional[FieldsOfExperts]=None,
                loss: Optional[torch.Tensor]=None, **kwargs) -> None:
        device = kwargs.get('device', None)
        dtype = kwargs.get('dtype', None)

        if step % self.plotting_freq == 0 and regulariser is not None:
            self._plot_filters(step, regulariser)
            self._plot_potentials(step, regulariser, device, dtype)

    @staticmethod
    def _normalise_filter(filter_tensor: torch.Tensor) -> torch.Tensor:
        filter_tensor = filter_tensor - torch.min(filter_tensor)
        filter_tensor = filter_tensor / torch.max(filter_tensor)
        return filter_tensor

    def _plot_filters(self, step: int, regulariser: FieldsOfExperts) -> None:
        filters = regulariser.get_image_filter().get_filter_tensor()
        filter_norms = [torch.linalg.norm(fltr).detach().cpu().item() for fltr in filters]

        filter_indices_sorted = np.argsort(filter_norms)[::-1].tolist()

        num_filters = filters.shape[0]
        num_filters_sqrt = int(math.sqrt(num_filters)) + 1
        fig, axes = plt.subplots(num_filters_sqrt, num_filters_sqrt, figsize=(11, 11),
                                 gridspec_kw={'hspace': 0.9, 'wspace': 0.2})

        for i in range(0, num_filters_sqrt):
            for j in range(0, num_filters_sqrt):
                filter_idx = i * num_filters_sqrt + j
                if filter_idx < num_filters:
                    idx = filter_indices_sorted[filter_idx]

                    fltr = self._normalise_filter(filters[idx, :, :, :].squeeze().detach().clone())
                    axes[i, j].imshow(fltr.cpu().numpy(), cmap=cmaps['gray'])

                    title = 'idx={:d}, \nnorm={:.3f}'.format(idx, filter_norms[idx])
                    axes[i, j].set_title(title, fontsize=8)
                    axes[i, j].axis('off')
                else:
                    fig.delaxes(axes[i, j])

        plt.savefig(os.path.join(self.path_to_filter_plot_dir, 'filters_iter_{:d}.png'.format(step)))
        if self.tb_writer:
            self.tb_writer.add_image('filters', figure_to_tensor(fig), step + 1)
        plt.close(fig)

    def _plot_potentials(self, step: int, regulariser: FieldsOfExperts, device: Optional[torch.device],
                         dtype: Optional[torch.dtype]) -> None:
        if device is not None and dtype is not None:
            image_filter = regulariser.get_image_filter()
            filter_tensors = image_filter.get_filter_tensor()
            filter_norms = [torch.linalg.norm(fltr).detach().cpu().item() for fltr in filter_tensors]
            filter_indices_sorted = np.argsort(filter_norms)[::-1].tolist()

            u_clean = list(self.test_loader)[0].to(dtype=dtype, device=device)
            filter_response = image_filter(u_clean)
            filter_response = torch.permute(filter_response, dims=(1, 0, 2, 3))
            filter_response_flat = torch.flatten(filter_response, start_dim=1, end_dim=3)
            quantiles = torch.quantile(filter_response_flat,
                                       q=torch.tensor([self.p_low, self.p_high]).to(device=device), dim=-1)

            potential = regulariser.get_potential()
            num_marginals = potential.get_num_marginals()
            num_marginals_sqrt = int(math.sqrt(num_marginals)) + 1

            t = torch.stack([torch.linspace(quantiles[0, i], quantiles[1, i], 111)
                             for i in range(0, num_marginals)]).unsqueeze(dim=1).unsqueeze(dim=0)
            t = t.to(device=device, dtype=dtype)
            rho = potential.forward(t, reduce=False)

            fig, axes = plt.subplots(num_marginals_sqrt, num_marginals_sqrt, figsize=(13, 13),
                                     gridspec_kw={'hspace': 0.9, 'wspace': 0.4}, sharex=False, sharey=False)
            for i in range(0, num_marginals_sqrt):
                for j in range(0, num_marginals_sqrt):
                    potential_idx = i * num_marginals_sqrt + j
                    if potential_idx < potential.get_num_marginals():
                        idx = filter_indices_sorted[potential_idx]

                        axes[i, j].plot(t[0, idx, 0, :].detach().cpu().numpy(),
                                        rho[0, idx, 0, :].detach().cpu().numpy() -
                                        torch.min(rho[0, potential_idx, 0, :]).detach().cpu().numpy(), color='blue')
                        if potential.__class__.__name__ == 'StudentT':
                            potential_weight_tensor = potential.get_parameters()
                            potential_weight = potential_weight_tensor[idx].detach().cpu().item()
                            axes[i, j].set_title('idx={:d}, \nweight={:.3f}'.format(idx, potential_weight),
                                                 fontsize=8)
                        else:
                            axes[i, j].set_title('idx={:d}'.format(idx), fontsize=8)

                        axes[i, j].set_xlim(quantiles[0, idx].cpu().item(), quantiles[1, idx].cpu().item())
                        for x_label, y_label in zip(axes[i, j].get_xticklabels(), axes[i, j].get_yticklabels()):
                            x_label.set_fontsize(8)
                            y_label.set_fontsize(8)
                    else:
                        fig.delaxes(axes[i, j])

            plt.savefig(os.path.join(self.path_to_potential_plot_dir, 'potentials_iter_{:d}.png'.format(step)))
            if self.tb_writer:
                self.tb_writer.add_image('potentials', figure_to_tensor(fig), step + 1)
            plt.close(fig)

    def on_train_begin(self, regulariser: Optional[FieldsOfExperts]=None, **kwargs) -> None:
        pass

    def on_train_end(self) -> None:
        pass

class TrainingMonitor(Callback):

    TENSORBOARD_TAGS = {'loss': 'loss/train', 'test_loss': 'loss/test', 'test_psnr': 'test_psnr'}
    HYPERPARAM_KEYS = {LIP_CONST_KEY, LR_KEY}

    def __init__(self, 
                 dataset: Dataset, 
                 method_lower: str, 
                 options_lower: Dict[str, Any],
                 loss_func: Callable, 
                 path_to_data_dir: str, 
                 operator: Optional[torch.nn.Module]=None, 
                 noise_level: Optional[float]=None,
                 lam: Optional[float]=None,
                 config: Optional[Configuration]=None,
                 evaluation_freq: int=2,
                 tb_writer: Optional[SummaryWriter]=None) -> None:
        super().__init__()
        self.test_loader = DataLoader(dataset, batch_size=len(dataset), shuffle=False,
                                 collate_fn=lambda x: collate_function(x, crop_size=-1))

        if config is not None:
            operator_name = config['measurement_model']['forward_operator'].get()
            self.operator = getattr(torch.nn, operator_name)()
            self.noise_level = config['measurement_model']['noise_level'].get()
            self.lam = config['energy']['lam'].get()
        elif noise_level is not None and operator is not None and lam is not None:
            self.operator = operator
            self.noise_level = noise_level
            self.lam = lam
        else:
            raise ValueError('Must provide config or, operator, noise_level and lam.')

        self.method_lower = method_lower
        self.options_lower = options_lower
        self.loss_func = loss_func

        self.path_to_data_dir = path_to_data_dir
        self.path_to_test_data_dir = os.path.join(self.path_to_data_dir, 'test')
        if not os.path.exists(self.path_to_test_data_dir):
            os.makedirs(self.path_to_test_data_dir, exist_ok=True)

        self.tb_writer = tb_writer

        self.evaluation_freq = evaluation_freq

        self.training_stats_dict_list = []
        self.potential_params_list = []

        self.hyperparam_dict = {}

    def on_train_begin(self, regulariser: Optional[FieldsOfExperts]=None, **kwargs) -> None:
        logging.info('[{:s}] compute initial test loss and initial psnr'.format(self.__class__.__name__))
        device = kwargs.get('device', None)
        dtype = kwargs.get('dtype', None)

        if regulariser is not None:
            self._evaluate_on_test_data(-1, regulariser, device, dtype)

    def on_step(self, step: int, regulariser: Optional[FieldsOfExperts]=None,
                loss: Optional[torch.Tensor]=None, **kwargs) -> None:
        if step % self.evaluation_freq == 0:
            logging.info('[{:s}] log statistics and hyperparameters'.format(self.__class__.__name__))

            param_groups = kwargs.get('param_groups', None)
            self._log_hyperparameters(step, param_groups)

            device = kwargs.get('device', None)
            dtype = kwargs.get('dtype', None)
            self._log_training_stats(step, regulariser, loss, device, dtype)

    def _log_hyperparameters(self, step: int, param_groups: Optional[List[Dict[str, Any]]]) -> None:
        if param_groups:
            for group in param_groups:
                name = group.get(PARAM_GROUP_NAME_KEY, '')
                if not name in self.hyperparam_dict.keys():
                    self.hyperparam_dict[name] = {}

                for key in self.HYPERPARAM_KEYS:
                    if key in group.keys():
                        hparam = group[key] if isinstance(group[key], float) else group[key][0]
                        if not key in self.hyperparam_dict[name]:
                            self.hyperparam_dict[name][key] = []

                        logging.info('[{:s}] {:s} for group {:s}: '
                                     '{:.3f}'.format(self.__class__.__name__, key, name, hparam))
                        self.hyperparam_dict[name][key].append(hparam)
                        if self.tb_writer:
                            self.tb_writer.add_scalar('{:s}/{:s}'.format(key, name), hparam, step + 1)

    def _log_training_stats(self, step: int, regulariser: Optional[FieldsOfExperts], loss: Optional[torch.Tensor],
                            device: Optional[torch.device], dtype: Optional[torch.dtype]) -> None:
        stats = {'step': step}
        if loss:
            loss_ = loss.detach().cpu().numpy()
            stats.update({'loss': loss_})
            if self.tb_writer:
                self.tb_writer.add_scalar(self.TENSORBOARD_TAGS['loss'], loss_, step + 1)

        if regulariser is not None:
            test_loss, test_psnr = self._evaluate_on_test_data(step, regulariser, device, dtype)

            for key, value in zip(['test_loss', 'test_psnr'], [test_loss, test_psnr]):
                if value:
                    stats.update({key: value})
                    if self.tb_writer:
                        self.tb_writer.add_scalar(self.TENSORBOARD_TAGS[key], value, step + 1)

            fitness = self._compute_model_fitness(loss, test_loss, test_psnr)
            if fitness:
                stats.update({'fitness': fitness})

            if self.tb_writer:
                potential_param = regulariser.potential.get_parameters().detach().cpu().numpy()
                if potential_param.ndim == 1:
                    potential_param_norms = np.exp(potential_param)
                else:
                    potential_param_norms = np.linalg.norm(np.exp(potential_param), 
                                                           axis=tuple(range(1, potential_param.ndim)))
                self.tb_writer.add_scalars('potentials/weight_norms',
                                           {'potential_{:d}'.format(i): potential_param_norms[i]
                                            for i in range(0, len(potential_param_norms))}, step + 1)

        self.training_stats_dict_list.append(stats)

    def on_train_end(self) -> None:
        df = pd.DataFrame.from_dict(self.training_stats_dict_list)
        train_loss_list = df['loss'].dropna().to_list()
        test_loss_list = df['test_loss'].dropna().to_list()
        test_psnr_list = df['test_psnr'].dropna().to_list()
        self._visualise_training_stats(train_loss_list, test_loss_list, test_psnr_list)

        self._visualise_hyperparam_stats()

        self._export_model_ranking(df[['step', 'fitness']].dropna())

    def _visualise_hyperparam_stats(self) -> None:
        num_param_groups = len(self.hyperparam_dict.keys())
        num_hparams = max(len(self.hyperparam_dict[param_name].keys()) 
            for param_name in self.hyperparam_dict)

        fig, axes = plt.subplots(num_param_groups, num_hparams, figsize=(7, 9), squeeze=False, gridspec_kw={"hspace": 0.5})

        for i, group_name in enumerate(self.hyperparam_dict.keys()):
            for j, hparam_key in enumerate(self.hyperparam_dict[group_name]):
                hparam_list = self.hyperparam_dict[group_name][hparam_key]

                axes[i, j].plot(self.evaluation_freq * np.arange(0, len(hparam_list)), hparam_list)
                axes[i, j].set_title('evolution of {:s}/{:s} for upper level problem'.format(group_name, hparam_key))
                axes[i, j].xaxis.get_major_locator().set_params(integer=True)
                axes[i, j].set_xlabel('iteration')

        plt.savefig(os.path.join(self.path_to_data_dir, 'hyperparam_stats.png'))
        plt.close(fig)

    def _visualise_training_stats(self, train_loss_list: List[float], test_loss_list: List[float],
                                  test_psnr_list: List[float]) -> None:

        moving_average = compute_moving_average(np.array(train_loss_list), 10)

        fig = plt.figure(figsize=(11, 11))

        ax_1 = fig.add_subplot(1, 2, 1)
        ax_1.set_title('training loss')
        ax_1.plot(self.evaluation_freq * np.arange(0, len(train_loss_list)), 
                  train_loss_list, label='train loss')
        ax_1.plot(self.evaluation_freq * np.arange(0, len(moving_average)), 
                  moving_average, color='orange', label='moving average of train loss')
        ax_1.plot(self.evaluation_freq * np.arange(0, len(test_loss_list)), test_loss_list,
                  color='cyan', label='test loss')
        ax_1.xaxis.get_major_locator().set_params(integer=True)
        ax_1.set_xlabel('iteration')
        ax_1.legend()

        ax_2 = fig.add_subplot(1, 2, 2)
        ax_2.set_title('average psnr over test set')
        ax_2.plot(self.evaluation_freq * np.arange(0, len(test_psnr_list)), test_psnr_list)
        ax_2.xaxis.get_major_locator().set_params(integer=True)
        ax_2.set_xlabel('iteration')

        plt.savefig(os.path.join(self.path_to_data_dir, 'training_stats.png'))
        plt.close(fig)

    def _evaluate_on_test_data(self, step: int, regulariser: FieldsOfExperts, device: Optional[torch.device],
                               dtype: Optional[torch.dtype]) -> Tuple[Optional[float], Optional[float]]:
        psnr = None
        loss_test = None
        if device is not None and dtype is not None:
            test_batch_clean = list(self.test_loader)[0]
            test_batch_clean_ = test_batch_clean.to(device=device, dtype=dtype)

            measurement_model = MeasurementModel(test_batch_clean_, self.operator, self.noise_level)
            energy = Energy(measurement_model, regulariser, self.lam)
            energy.to(device=device, dtype=dtype)

            test_batch_noisy = measurement_model.get_noisy_observation()
            with Timer(device) as t:
                test_batch_denoised = solve_lower(energy, self.method_lower, self.options_lower).solution

            psnr = torch.mean(compute_psnr(energy.measurement_model.get_clean_data(), test_batch_denoised))
            psnr = psnr.detach().cpu().item()
            loss_test = self.loss_func(test_batch_clean_, test_batch_denoised)
            loss_test = loss_test.detach().cpu().item()

            logging.info('[{:s}]   > average psnr: {:.5f}'.format(self.__class__.__name__, psnr))
            logging.info('[{:s}]   > test loss: {:.5f}'.format(self.__class__.__name__, loss_test))
            logging.info('[{:s}]   > evaluation took [ms]: {:.5f}'.format(self.__class__.__name__, t.time_delta()))

            u_clean_splits = torch.split(test_batch_clean_, split_size_or_sections=1, dim=0)
            u_noisy_splits = torch.split(test_batch_noisy, split_size_or_sections=1, dim=0)
            u_denoised_splits = torch.split(test_batch_denoised, split_size_or_sections=1, dim=0)

            num_rows = len(u_clean_splits)
            num_cols = 3
            fig, axes = plt.subplots(num_rows, num_cols, figsize=(11, 11),
                                     gridspec_kw={'hspace': 0.9, 'wspace': 0.2}, sharex=True, sharey=True)
            if num_rows == 1:
                axes = [axes]
            for idx, (item_clean, item_noisy, item_denoised) in (
                    enumerate(zip(u_clean_splits, u_noisy_splits, u_denoised_splits))):

                axes[idx][0].imshow(item_clean.squeeze().detach().cpu().numpy(), cmap=cmaps['gray'])
                axes[idx][1].imshow(item_noisy.squeeze().detach().cpu().numpy(), cmap=cmaps['gray'])
                axes[idx][2].imshow(item_denoised.squeeze().detach().cpu().numpy(), cmap=cmaps['gray'])

                if idx == 0:
                    axes[idx][0].set_title('clean')
                    axes[idx][1].set_title('noisy')
                    axes[idx][2].set_title('denoised')

                axes[idx][0].axis("off")
                axes[idx][1].axis("off")
                axes[idx][2].axis("off")

            plt.savefig(os.path.join(self.path_to_test_data_dir, 'triplet_iter_{:d}.png'.format(step)))
            if self.tb_writer:
                self.tb_writer.add_image('triplets/test', figure_to_tensor(fig), step + 1)
            plt.close(fig)

        return loss_test, psnr

    def _compute_model_fitness(self, train_loss: Optional[float], test_loss: Optional[float],
                               test_psnr: Optional[float], **kwargs: Optional[float]) -> Optional[float]:
        """
        Function which computes the so-called fitness of the current model. The fitness is intended to introduce a
        model ranking.

        NOTE
        ----
            > Currently, fitness is simply the psnr.

        :param train_loss: Current training loss
        :param test_loss: Current test loss
        :param test_psnr: Current psnr
        :param kwargs: Additional keyword arguments, which may be used when generalising, or updating the
            definition of fitness.
        :return: Float representing the fitness value
        """
        fitness = None
        if not any(item is None for item in [train_loss, test_loss, test_psnr]):
            fitness = 0.0 * train_loss + 0.0 * test_loss + 1.0 * test_psnr
        return fitness

    def _export_model_ranking(self, fitness: pd.DataFrame) -> None:
        fitness_sorted = fitness.sort_values('fitness', ascending=False, na_position='last', ignore_index=True)
        fitness_sorted.index += 1
        fitness_sorted.head(10).to_csv(os.path.join(self.path_to_data_dir, 'model_ranking.csv'))


from typing import Dict, Any
import torch
import logging
import numpy as np

from pylopt.regularisers.fields_of_experts import FieldsOfExperts
from pylopt.utils.Timer import Timer

def warm_up(regulariser: FieldsOfExperts, x: torch.Tensor, num_calls: int = 100) -> None:
    logging.info('[WARM UP] Perform warm-up forward and backward calls after compilation.')
    x_ = x.detach().clone().requires_grad_(True)

    _ = regulariser(x_)

    timings = []
    for _ in range(0, num_calls):
        with Timer(x.device) as t:
            with torch.enable_grad():
                y = regulariser(x_)
            _ = torch.autograd.grad(inputs=x_, outputs=y)
        timings.append(t.time_delta())

    logging.info('[WARM UP] mean elapsed time [ms]: {:.5f}'.format(np.mean(timings)))
    logging.info('[WARM UP] median elapsed time [ms]: {:.5f}'.format(np.median(timings)))

def compile_regulariser(regulariser: FieldsOfExperts, x: torch.Tensor,
                        **compile_options: Dict[str, Any]) -> FieldsOfExperts:
    """
    This function compiles the regulariser using torch.compile() for improved memory management and
    speed-up in forward and backward calls. After compilation a few warm-up iterations were performed.

    NOTE
    ----
        > **IMPORTANT** Call this method only after assigning the module to a device:

            ...
            energy = Energy(...)
            energy.to(device=device)
            energy.compile_regulariser(...)
            ...

        > Compilation is not only suitable for GPU-based inference or training - also models on CPU can
            be compiled.
        > Compilation takes some time in the beginning - this is why warm-up forward and backward calls
            were performed right after compilation.
        > Per default the most aggressive optimisation options were applied. For all the available options see
            https://docs.pytorch.org/tutorials/intermediate/torch_compile_tutorial.html

    :param regulariser: Instance of FieldsOfExperts to be compiled.
    :param x: PyTorch tensor which is used in the warm-up procedure to perform several forward and backward calls.
    :param compile_options: Optional dictionary of compiling options - for the available options see
        https://docs.pytorch.org/tutorials/intermediate/torch_compile_tutorial.html
    :return: compiled FieldsOfExperts model
    """

    torch._functorch.config.donated_buffer = False

    mode = compile_options.get('mode', 'max-autotune')
    backend = compile_options.get('backend', 'inductor')
    dynamic = compile_options.get('dynamic', True)
    fullgraph = compile_options.get('fullgraph', True)

    regulariser = torch.compile(regulariser, mode=mode, backend=backend, dynamic=dynamic, fullgraph=fullgraph)
    warm_up(regulariser, x, num_calls=100)
    return regulariser



from typing import Callable, Any, Optional, Dict, List, Tuple
import torch
from torch.utils.data import DataLoader
import logging
from confuse import Configuration
from itertools import chain

from pylopt.bilevel_problem.gradients import ImplicitAutogradFunction, UnrollingAutogradFunction
from pylopt.bilevel_problem.parameter_groups import get_param_group_name, PARAM_GROUP_NAME_KEY
from pylopt.callbacks import Callback
from pylopt.energy.Energy import Energy
from pylopt.energy.MeasurementModel import MeasurementModel
from pylopt.optimise import step_adam, create_projected_optimiser, step_nag
from pylopt.optimise.optimise_adam import harmonise_param_groups_adam
from pylopt.optimise.optimise_lbfgs import harmonise_param_groups_lbfgs
from pylopt.optimise.optimise_nag import harmonise_param_groups_nag
from pylopt.regularisers.fields_of_experts.FieldsOfExperts import FieldsOfExperts
from pylopt.regularisers.fields_of_experts.ImageFilter import ImageFilter
from pylopt.regularisers.fields_of_experts.potential.Potential import Potential
from pylopt.scheduler import HyperParamScheduler
from pylopt.solver.CGSolver import CGSolver
from pylopt.dataset.dataset_utils import collate_function
from pylopt.utils.file_system_utils import create_experiment_dir, dump_configs, \
    dump_bilevel_training_settings

def assemble_param_groups_base(regulariser: FieldsOfExperts, alternating: bool=False):
    param_dict = {}
    for child in regulariser.children():
        key = get_param_group_name(child)
        if key is None:
            raise ValueError('Regulariser is supposed to have only children of '
                             'type {:s}, {:s}'.format(ImageFilter.__name__, Potential.__name__))
        param_dict[key] = [p for p in child.parameters() if p.requires_grad]

    param_groups = []
    if not alternating:
        group = {'params': list(chain.from_iterable([param_list for param_list in param_dict.values()])),
                 PARAM_GROUP_NAME_KEY: 'joint'}
        param_groups.append(group)
    else:
        for key, value in param_dict.items():
            if value:
                group = {'params': value, PARAM_GROUP_NAME_KEY: key}
                param_groups.append(group)
    return param_groups

def assemble_param_groups_adam(regulariser: FieldsOfExperts, lr: Optional[List[float]]=None,
                               betas: Optional[List[Tuple[float, float]]]=None, eps: Optional[List[float]]=None,
                               weight_decay: Optional[List[float]]=None,
                               parameterwise: bool=True, **unknown_options) -> List[Dict[str, Any]]:
    param_groups = assemble_param_groups_base(regulariser, parameterwise)

    lr = [None for _ in range(0, len(param_groups))] if not lr else lr
    betas = [None for _ in range(0, len(param_groups))] if not betas else betas
    eps = [None for _ in range(0, len(param_groups))] if not eps else eps
    weight_decay = [None for _ in range(0, len(param_groups))] if not weight_decay else weight_decay

    for idx, group in enumerate(param_groups):
        group['lr'] = lr[idx]
        group['betas'] = betas[idx]
        group['eps'] = eps[idx]
        group['weight_decay'] = weight_decay[idx]

    return param_groups

def assemble_param_groups_nag(regulariser: FieldsOfExperts, alpha: Optional[List[float]]=None,
                              beta: Optional[List[float]]=None,
                              lip_const: Optional[List[float]]=None,
                              alternating: bool=True, **unknown_options) -> List[Dict[str, Any]]:
    param_groups = assemble_param_groups_base(regulariser, alternating)

    alpha = [None for _ in range(0, len(param_groups))] if not alpha else alpha
    beta = [None for _ in range(0, len(param_groups))] if not beta else beta
    lip_const = [None for _ in range(0, len(param_groups))] if not lip_const else lip_const

    for idx, group in enumerate(param_groups):
        group['alpha'] = [alpha[idx]]
        group['beta'] = [beta[idx]]
        group['lip_const'] = [lip_const[idx]]

    return param_groups

def assemble_param_groups_lbfgs(regulariser: FieldsOfExperts, lr: Optional[List[float]]=None,
                                max_iter: Optional[int]=None,
                                max_eval: Optional[int]=None, tolerance_grad: Optional[float]=None,
                                tolerance_change: Optional[float]=None, history_size: Optional[int]=None,
                                line_search_fn: Optional[str]=None, **unknown_options) -> List[Dict[str, Any]]:
    # NOTE
    #   > According to
    #           https://docs.pytorch.org/docs/stable/generated/torch.optim.LBFGS.html,
    #       the PyTorch implementation currently supports only a single parameter group.

    param_groups = assemble_param_groups_base(regulariser, False)
    for group in param_groups:
        group['lr'] = None if not lr else lr
        group['max_iter'] = None if not max_iter else max_iter
        group['max_eval'] = None if not max_eval else max_eval
        group['tolerance_grad'] = None if not tolerance_grad else tolerance_grad
        group['tolerance_change'] = None if not tolerance_change else tolerance_change
        group['history_size'] = None if not history_size else history_size
        group['line_search_fn'] = None if not line_search_fn else line_search_fn

    return param_groups

class BilevelOptimisation:
    """
    This class implements the routines for the training of filters and/or potentials of a FoE model. For the training
    gradient based methods are considered, where gradients are computed using implicit differentiation or
    an unrolling scheme. The unrolling scheme requires that the lower level problem is solved using NAG or NAPG.
    """
    def __init__(self,
                 method_lower: str, 
                 options_lower: Dict[str, Any], 
                 operator: Optional[torch.nn.Module]=None,
                 noise_level: Optional[float]=None,
                 config: Optional[Configuration]=None,
                 differentiation_method: str='implicit', 
                 solver_name: Optional[str]='cg',
                 options_solver: Optional[Dict[str, Any]]=None,
                 path_to_experiments_dir: Optional[str]=None
    ) -> None:
        """
        Initialisation of BilevelOptimisation.

        NOTE
        ----
            > For the setup of an instance of this class or the operator and the noise level 
                must be specified directly when calling the constructor, or by means of 
                a configuration object of type Configuration.
            > If a configuation object is provided, the config values are written to file.

        TODO
        ----
            > Make class completely configurable by config.

        :param method_lower: String indicating the solution method for the lower level problem.
        :param options_lower: Dictionary of options for the solution of the lower level problem
        :param operator: PyTorch module representing the forward operator
        :param noise_level: Float in the interval (0, 1) representing the noise level.
        :param config: Configuration object for the setup of measurement model and energy
        :param differentiation_method: String indicating which differentiation method is used:
            > 'implicit': Computation of derivative of upper level objective function using the implicit
                function theorem. This approach requires to solve a linear system of equations - hence a linear
                system solver must be specified if choosing this option.
            > 'hessian_free': Derivative of upper level objective function is computed by approximating
                the system matrix (involving the hessian of the energy function) of the implicit differentiation
                scheme using the identity matrix.
            > 'unrolling': This scheme is about performing a small number of gradient steps to solve the lower level
                problem while not breaking the corresponding computational graph. Derivatives of the resulting
                approximate solution of the lower level problem are then computed by means of autograd.
            Per default, implicit differentiation is applied.
        :param solver_name: String indicating the linear system solver which is used to compute gradients when implicit
            differentiation is applied. The default option is 'cg'.
        :param options_solver: Dictionary of options for the linear system solver. The cg solver takes the
            options
                max_num_iterations: int, optional
                    Maximal number of cg iterations; the default value equals 500.
                abs_tol: float, optional
                    Tolerance used for early stopping: Iteration is stopped if the cg residual is <= abs_tol. Per
                    default the tolerance value 1e-5 is used.
        :param path_to_experiments_dir: Path to experiment directory where intermediate results, optimisation stats, ecc.
            will be stored. If not provided, an experiment directory in the root directory of this python
            project is created.
        """
        self.path_to_experiments_dir = create_experiment_dir(self.config) if path_to_experiments_dir is None \
            else path_to_experiments_dir

        self.backward_mode = differentiation_method

        self.method_lower = method_lower
        self.options_lower = options_lower

        self.config = config
        if self.config is not None:
            dump_configs(config, self.path_to_experiments_dir)
        self.noise_level = noise_level
        self.operator = operator

        self.solver_name = solver_name
        if self.solver_name == 'cg':
            self.options_solver = options_solver if options_solver is not None else {}
            self.solver = CGSolver(**self.options_solver)
        elif self.solver_name == 'your_custom_solver':
            # Custom linear system solver goes here. Use the following structure
            #
            #    self.options_solver = ...
            #    self.solver = MySolver(**self.options_solver)
            pass
        else:
            raise ValueError('Unknown linear system solver')

    def _loss_func_factory(self, upper_loss: Callable, energy: Energy, u_clean: torch.Tensor) -> Callable:
        if self.backward_mode == 'unrolling':
            def func(*params: torch.nn.Parameter) -> torch.Tensor:
                upper_loss_func = lambda z: upper_loss(u_clean, z)
                return UnrollingAutogradFunction.apply(energy, self.method_lower, self.options_lower,
                                                       upper_loss_func, *params)
        else:
            def func(*params: torch.nn.Parameter) -> torch.Tensor:
                upper_loss_func = lambda z: upper_loss(u_clean, z)
                return ImplicitAutogradFunction.apply(energy, self.method_lower, self.options_lower,
                                                      upper_loss_func, self.solver,
                                                      self.backward_mode == 'hessian_free',
                                                      *params)
        return func

    def learn(self, regulariser: FieldsOfExperts, lam: float, upper_loss_func: Callable, dataset_train,
              optimisation_method_upper: str, optimisation_options_upper: Dict[str, Any],
              batch_size: int=32, crop_size: int=64, dtype: torch.dtype=torch.float32,
              device: Optional[torch.device]=None, callbacks: Optional[List[Callback]]=None,
              schedulers: Optional[List[HyperParamScheduler]]=None, do_compile: bool=False) -> None:

        upper_settings = {'method': optimisation_method_upper, 'options': optimisation_options_upper}
        options_lower_ = dict(self.options_lower)
        options_lower_.update({'prox': True if 'prox' in self.options_lower.keys() else False})
        lower_settings = {'method': self.method_lower, 'options': options_lower_}
        backward_settings = {'method': self.backward_mode}
        if self.backward_mode:
            backward_settings.update({'options': {'solver': self.solver_name, 'solver_options': self.options_solver}})
        dump_bilevel_training_settings(upper_settings, lower_settings, backward_settings, self.path_to_experiments_dir)

        device = device if device is not None else torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        train_loader = DataLoader(dataset_train, batch_size=batch_size,
                                  collate_fn=lambda x: collate_function(x, crop_size=crop_size))

        if optimisation_method_upper == 'nag':
            self._learn_nag(regulariser, lam, upper_loss_func, optimisation_options_upper, train_loader,
                            dtype, device, do_compile, callbacks, schedulers)
        elif optimisation_method_upper == 'adam':
            self._learn_adam(regulariser, lam, upper_loss_func, optimisation_options_upper, train_loader,
                            dtype, device, callbacks, schedulers)
        elif optimisation_method_upper == 'lbfgs':
            self._learn_lbfgs(regulariser, lam, upper_loss_func, optimisation_options_upper, train_loader,
                            dtype, device, callbacks, schedulers)
        elif optimisation_method_upper == 'your_custom_method':
            pass
        else:
            raise ValueError('Unknown optimisation method for upper level problem.')

    @torch.no_grad()
    def _learn_nag(self, regulariser: FieldsOfExperts, lam: float, upper_loss: Callable,
                   optimisation_options_upper: Dict[str, Any], train_loader: torch.utils.data.DataLoader,
                   dtype: torch.dtype, device: torch.device, do_compile: bool,
                   callbacks: Optional[List[Callback]]=None,
                   schedulers: Optional[List[HyperParamScheduler]]=None) -> None:
       if callbacks is None:
           callbacks = []
       if schedulers is None:
           schedulers = []

       regulariser = regulariser.to(device=device, dtype=dtype)
       # if do_compile:
       #     train_iter = iter(train_loader)
       #     batch_0 = next(train_iter).to(dtype=dtype, device=device)
       #     # NOTE:
       #     #    > Use default compilation options here - in particular the dynamic shape option.
       #     #    > Dynamic shapes are important here since the same model is used for training and testing (on full
       #     #        images!)
       #     regulariser = compile_regulariser(regulariser, batch_0, dynamic=True)

       # DEBUG:begin
       # train_iter = iter(train_loader)
       # batch_0 = next(train_iter).to(dtype=dtype, device=device)
       # with torch.enable_grad():
       #     x = batch_0.detach().clone().requires_grad_(True).to(dtype=dtype, device=device)
       #     y = regulariser.forward(x)
       #     dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)[0]
       #     grad_nodal = torch.autograd.grad(inputs=[p for p in regulariser.parameters() if p.requires_grad], outputs=y, create_graph=True)[0]
       #
       # dy_dx_ = (first_derivative..)
       #
       # -> mixed derivative does not exist?? why?
       #
       # d2y_dx2 = torch.autograd.grad(inputs=[p for p in regulariser.parameters() if p.requires_grad], outputs=dy_dx, grad_outputs=torch.rand_like(x))
       #
       # asdf
       # DEBUG:end


       param_groups = assemble_param_groups_nag(regulariser, **optimisation_options_upper)
       param_groups_ = harmonise_param_groups_nag(param_groups)

       for sched in schedulers:
           sched.bind(param_groups_)

       for cb in callbacks:
           cb.on_train_begin(regulariser, device=device, dtype=dtype)

       max_num_iterations = optimisation_options_upper['max_num_iterations']

       try:
           for k, batch in enumerate(train_loader):
               batch_ = batch.to(dtype=dtype, device=device)
               measurement_model = MeasurementModel(batch_, config=self.config, 
                                                    operator=self.operator, noise_level=self.noise_level)

               energy = Energy(measurement_model, regulariser, lam)
               energy = energy.to(device=device, dtype=dtype)

               func = self._loss_func_factory(upper_loss, energy, batch_)
               def grad_func(*params):
                   with torch.enable_grad():
                       loss_ = func(*params)
                   return list(torch.autograd.grad(outputs=loss_, inputs=params))
               loss = step_nag(func, grad_func, param_groups_)

               for cb in callbacks:
                   cb.on_step(k + 1, regulariser, loss, param_groups=param_groups_, device=device, dtype=dtype)

               for sched in schedulers:
                   sched.step(func=func, grad_func=grad_func)

               logging.info('[TRAIN] iteration [{:d} / {:d}]: '
                            'loss = {:.5f}'.format(k + 1, max_num_iterations, loss.detach().cpu().item()))

               if (k + 1) == max_num_iterations:
                   logging.info('[TRAIN] reached maximal number of iterations')
                   break
       finally:
           for cb in callbacks:
               cb.on_train_end()

    def _learn_adam(self, regulariser: FieldsOfExperts, lam: float, upper_loss: Callable,
                    optimisation_options_upper: Dict[str, Any], train_loader: torch.utils.data.DataLoader,
                    dtype: torch.dtype, device: torch.device, callbacks: Optional[List[Callback]]=None,
                    schedulers: Optional[List[HyperParamScheduler]]=None) -> None:
        if callbacks is None:
            callbacks = []
        if schedulers is None:
            schedulers = []

        param_groups = assemble_param_groups_adam(regulariser, **optimisation_options_upper)
        param_groups_ = harmonise_param_groups_adam(param_groups)

        for sched in schedulers:
            sched.bind(param_groups_)

        for cb in callbacks:
            cb.on_train_begin(regulariser, device=device, dtype=dtype)

        optimiser = create_projected_optimiser(torch.optim.Adam)(param_groups_)
        max_num_iterations = optimisation_options_upper['max_num_iterations']
        try:
            for k, batch in enumerate(train_loader):
                with torch.no_grad():
                    batch_ = batch.to(dtype=dtype, device=device)

                    measurement_model = MeasurementModel(batch_, config=self.config,
                                                         operator=self.operator, noise_level=self.noise_level)
                    energy = Energy(measurement_model, regulariser, lam)
                    energy = energy.to(device=device, dtype=dtype)

                    func = self._loss_func_factory(upper_loss, energy, batch_)
                    loss = step_adam(optimiser, func, param_groups_)

                    def grad_func(*params):
                        with torch.enable_grad():
                            loss_ = func(*params)
                        return list(torch.autograd.grad(outputs=loss_, inputs=params))
                    for sched in schedulers:
                        sched.step(func=func, grad_func=grad_func)

                    for cb in callbacks:
                        cb.on_step(k + 1, regulariser, loss, param_groups=param_groups_, device=device, dtype=dtype)

                    logging.info('[TRAIN] iteration [{:d} / {:d}]: '
                                 'loss = {:.5f}'.format(k + 1, max_num_iterations, loss.detach().cpu().item()))

                if (k + 1) == max_num_iterations:
                    logging.info('[TRAIN] reached maximal number of iterations')
                    break
        finally:
            for cb in callbacks:
                cb.on_train_end()

    def _learn_lbfgs(self, regulariser: FieldsOfExperts, lam: float, upper_loss: Callable,
                    optimisation_options_upper: Dict[str, Any], train_loader: torch.utils.data.DataLoader,
                    dtype: torch.dtype, device: torch.device, callbacks: Optional[List[Callback]]=None,
                    schedulers: Optional[List[HyperParamScheduler]]=None) -> None:
        if callbacks is None:
            callbacks = []
        if schedulers is None:
            schedulers = []

        param_groups_ = assemble_param_groups_lbfgs(regulariser, **optimisation_options_upper)
        param_groups_ = harmonise_param_groups_lbfgs(param_groups_)
        optimiser = create_projected_optimiser(torch.optim.LBFGS)(param_groups_)

        for sched in schedulers:
            sched.bind(param_groups_)

        for cb in callbacks:
            cb.on_train_begin(regulariser, device=device, dtype=dtype)

        max_num_iterations = optimisation_options_upper['max_num_iterations']
        try:
            for k, batch in enumerate(train_loader):
                with torch.no_grad():
                    batch_ = batch.to(dtype=dtype, device=device)

                    measurement_model = MeasurementModel(batch_, config=self.config,
                                                         operator=self.operator, noise_level=self.noise_level)
                    energy = Energy(measurement_model, regulariser, lam)
                    energy = energy.to(device=device, dtype=dtype)

                    func = self._loss_func_factory(upper_loss, energy, batch_)
                    def closure():
                        optimiser.zero_grad()
                        with torch.enable_grad():
                            loss = func(*[p for p in energy.parameters() if p.requires_grad])
                            loss.backward()
                        return loss
                    loss = optimiser.step(closure)

                    for cb in callbacks:
                        cb.on_step(k + 1, regulariser, torch.Tensor(loss), param_groups=param_groups_,
                                   device=device, dtype=dtype)

                    def grad_func(*params):
                        with torch.enable_grad():
                            loss_ = func(*params)
                        return list(torch.autograd.grad(outputs=loss_, inputs=params))
                    for sched in schedulers:
                        sched.step(func=func, grad_func=grad_func)

                    logging.info('[TRAIN] iteration [{:d} / {:d}]: '
                                 'loss = {:.5f}'.format(k + 1, max_num_iterations, loss))

                    if (k + 1) == max_num_iterations:
                        logging.info('[TRAIN] reached maximal number of iterations')
                        break
        finally:
            for cb in callbacks:
                cb.on_train_end()


from typing import Any, Tuple
import torch
from matplotlib import pyplot as plt
import numpy as np

from bilevel_optimisation.utils.Timer import Timer
#
# SUPP_LOWER = -2.5
# SUPP_UPPER = 2.5
# NUM_NODES = 6

def forward_func(x_scaled, coeffs, weights):
    y = (coeffs[..., 0]
         + x_scaled * (coeffs[..., 1]
                       + x_scaled * (coeffs[..., 2] +
                                     x_scaled * (coeffs[..., 3] + x_scaled * coeffs[..., 4]))))
    # return torch.sum(y * weights, dim=2)
    # out = (y * w).sum(dim=2)
    # return torch.einsum('bfnhw, n->bfhw', y, weights)
    return y * weights.view(1, 1, -1, 1, 1)

def backward_func(x_scaled, b, c, d, e, weights):
    y = (b + x_scaled * (2 * c + x_scaled * (3 * d + x_scaled * 4 * e)))
    return y * weights.view(1, 1, -1, 1, 1)
    # return torch.einsum('bfnhw, n->bfhw', y, weights)

class QuarticBSplineFunction(torch.autograd.Function):

    # @staticmethod
    # def _bucketise(x):
    #     x_scaled = ((x - SUPP_LOWER) /
    #                 (SUPP_UPPER - SUPP_LOWER))
    #     return torch.clamp((x_scaled * (NUM_NODES - 1)).ceil().long(),
    #                        min=0, max=NUM_NODES)

    @staticmethod
    def forward(x_scaled, scale, coeffs, weights) -> Any:
        # x_scaled = x.unsqueeze(dim=2)
        # x_scaled = (x_scaled - centers) / scale
        # index_tensor = QuarticBSplineFunction._bucketise(x_scaled)
        #
        # coeffs_indexed = coeffs[index_tensor]


        # b = coeffs[..., 1]
        # c = coeffs[..., 2]
        # d = coeffs[..., 3]
        # e = coeffs[..., 4]
        #
        # y = (coeffs[..., 0]
        #      + x_scaled * (b
        #                    + x_scaled * (c +
        #                                  x_scaled * (d + x_scaled * e))))
        #
        # return torch.einsum('bfnhw, n->bfhw', y, weights), b, c, d, e
        return forward_func(x_scaled, coeffs, weights), coeffs[..., 1], coeffs[..., 2], coeffs[..., 3], coeffs[..., 4]

    @staticmethod
    def setup_context(ctx: torch.autograd.function.FunctionCtx, inputs: Tuple, outputs: Tuple):
        x_scaled, scale, _, weights = inputs
        _, b, c, d, e = outputs
        ctx.save_for_backward(x_scaled, weights, b, c, d, e)
        ctx.scale = scale

    @staticmethod
    def backward(ctx: torch.autograd.function.FunctionCtx, *grad_outputs: Any) -> Any:
        x_scaled, weights, b, c, d, e = ctx.saved_tensors
        scale = ctx.scale

        # y = (b + x_scaled * (2 * c + x_scaled * (3 * d + x_scaled * 4 * e)))
        y = backward_func(x_scaled, b, c, d, e, weights / scale)
        return grad_outputs[0] * y, *[None] * 3
        # return grad_outputs[0] * torch.sum(y * weights / scale, dim=2), *[None] * 4

class QuarticBSplinePotential(torch.nn.Module):

    SUPP_LOWER = -2.5
    SUPP_UPPER = 2.5
    NUM_NODES = 6

    def __init__(self):
        super().__init__()

        coeffs = torch.zeros(7, 5)
        coeffs[1, :] = torch.tensor([625 / 16, 125 / 2, 75 / 2, 10.0, 1.0]) / 24
        coeffs[2, :] = torch.tensor([55 / 4, -5.0, -30.0, -20.0, -4.0]) / 24.0
        coeffs[3, :] = torch.tensor([115 / 8, 0.0, -15.0, 0.0, 6]) / 24.0
        coeffs[4, :] = torch.tensor([55 / 4, 5.0, -30.0, 20.0, -4.0]) / 24.0
        coeffs[5, :] = torch.tensor([625 / 16, -125 / 2, 75 / 2, -10.0, 1.0]) / 24
        self.register_buffer('coeffs', coeffs)

        box_lower = -3
        box_upper = 3
        num_centers = 33

        centers = torch.linspace(box_lower, box_upper, num_centers)
        self.register_buffer('centers', centers.view(1, 1, -1, 1, 1))
        self.weights = torch.nn.Parameter(torch.log(1 + centers ** 2), requires_grad=True)

        self.scale = (box_upper - box_lower) / (num_centers - 1)

    @classmethod
    def _bucketise(cls, x):
        x_scaled = (x - cls.SUPP_LOWER) / (cls.SUPP_UPPER - cls.SUPP_LOWER)
        return torch.clamp((x_scaled * (cls.NUM_NODES - 1)).ceil().long(), min=0, max=cls.NUM_NODES)

    def forward_(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:

        x_scaled = x.unsqueeze(dim=2)
        x_scaled = (x_scaled - self.centers) / self.scale
        index_tensor = self._bucketise(x_scaled)

        coeffs = self.coeffs[index_tensor]

        y = (coeffs[..., 0]
             + x_scaled * (coeffs[..., 1]
                           + x_scaled * (coeffs[..., 2] +
                                         x_scaled * (coeffs[..., 3] + x_scaled * coeffs[..., 4]))))

        y = torch.einsum('bfnhw, n->bfhw', y, self.weights)
        return torch.sum(y) if reduce else y

    def forward(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:
        x_scaled = x.unsqueeze(dim=2)
        x_scaled = (x_scaled - self.centers) / self.scale
        index_tensor = self._bucketise(x_scaled)

        coeffs = self.coeffs[index_tensor]

        # y = (coeffs[..., 0]
        #      + x_scaled * (coeffs[..., 1]
        #                    + x_scaled * (coeffs[..., 2] +
        #                                  x_scaled * (coeffs[..., 3] + x_scaled * coeffs[..., 4]))))
        #
        # y = torch.einsum('bfnhw, n->bfhw', y, self.weights)


        y, *_ = QuarticBSplineFunction.apply(x_scaled, self.scale, coeffs, self.weights)


        return torch.sum(y) if reduce else torch.sum(y, dim=(2))


def main():
    device = torch.device('cuda:0')

    pot = QuarticBSplinePotential().to(device=device)

    x = torch.rand(10, 48, 25, 25, device=device, requires_grad=True).to(device=device)
    y = pot(x)
    # y_ = pot.forward_(x)
    # dy_dx = torch.autograd.grad(inputs=x, outputs=y)
    # dy_dx_ = torch.autograd.grad(inputs=x, outputs=y_)

    # pot = torch.compile(pot, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)
    #
    # num_predictions = 1000
    # x = torch.rand(10, 48, 256, 256, device=device, requires_grad=True).to(device=device)
    #
    # # profile(pot, device, True)
    #
    # timings = []
    # for _ in range(0, num_predictions):
    #     with Timer(device) as t:
    #         y = pot(x)
    #         with torch.no_grad():
    #             dy_dx = torch.autograd.grad(inputs=x, outputs=y)
    #
    #     timings.append(t.time_delta())
    #
    # print('median: {:.5f}'.format(np.median(timings)))
    # print('mean: {:.5f}'.format(np.mean(timings)))
    # print('mean (tail): {:.5f}'.format(np.mean(timings[5::])))

    t = torch.stack([torch.linspace(-2, 2, 111)
                     for _ in range(0, 48)]).unsqueeze(dim=1).unsqueeze(dim=0).requires_grad_(True).to(device=device)

    p = pot(t, reduce=False)

    fig = plt.figure()
    ax_1 = fig.add_subplot(3, 1, 1)
    ax_2 = fig.add_subplot(3, 1, 2)
    ax_3 = fig.add_subplot(3, 1, 3)


    ax_1.plot(t[0, 0, 0, :].detach().cpu().numpy(), p[0, 0, 0, :].detach().cpu().numpy(), color='orange')
    ax_2.plot(t[0, 0, 0, :].detach().cpu().numpy(), p[0, 1, 0, :].detach().cpu().numpy(), color='orange')
    ax_3.plot(t[0, 0, 0, :].detach().cpu().numpy(), p[0, 2, 0, :].detach().cpu().numpy(), color='orange')
    plt.show()

def b_splines():



    x = torch.linspace(0.0, 1.0, 3000)

    nodes = torch.tensor([0.0, 0.25, 0.5, 0.75, 1.0])
    l = len(nodes) - 2
    kk = 4
    n = l + kk
    print(n)

    nodes_extended = torch.cat([nodes[0] * torch.ones(kk - 1), nodes, nodes[-1] * torch.ones(kk - 1)])
    basis = torch.zeros((len(x), len(nodes_extended) - 1))

    # if kk == 1:
    #     for j in range(0, n - k):
    #         basis[:, j] = (x >= nodes_extended[j]) & (x < nodes_extended[j + 1])

    if kk >= 2:
        delta = 0.25
        k = 1
        # for j in range(0, len(nodes_extended) - k - 1 - 1):
        for j in range(0, len(nodes_extended) - k):
            basis[:, j] = (x >= nodes_extended[j]) & (x < nodes_extended[j + 1])

        y = x.reshape(-1, 1) - nodes_extended.reshape(1, -1)

        k = 2
        for j in range(0, len(nodes_extended) - k):
            if nodes_extended[j] == nodes_extended[j + k - 1]:
                left = 0.0
            else:
                left = ((x - nodes_extended[j]) / (nodes_extended[j + k - 1] - nodes_extended[j])) * basis[:, j]

            if nodes_extended[j + 1] == nodes_extended[j + k]:
                right = 0.0
            else:
                right = ((nodes_extended[j + k] - x) / (nodes_extended[j + k] - nodes_extended[j + 1])) * basis[:, j + 1]
            basis[:, j] = left + right

        k = 3
        for j in range(0, len(nodes_extended) - k):
            if nodes_extended[j] == nodes_extended[j + k - 1]:
                left = 0.0
            else:
                left = ((x - nodes_extended[j]) / (nodes_extended[j + k - 1] - nodes_extended[j])) * basis[:, j]

            if nodes_extended[j + 1] == nodes_extended[j + k]:
                right = 0.0
            else:
                right = ((nodes_extended[j + k] - x) / (nodes_extended[j + k] - nodes_extended[j + 1])) * basis[:, j + 1]
            basis[:, j] = left + right

        k = 4
        for j in range(0, len(nodes_extended) - k):
            if nodes_extended[j] == nodes_extended[j + k - 1]:
                left = 0.0
            else:
                left = ((x - nodes_extended[j]) / (nodes_extended[j + k - 1] - nodes_extended[j])) * basis[:, j]

            if nodes_extended[j + 1] == nodes_extended[j + k]:
                right = 0.0
            else:
                right = ((nodes_extended[j + k] - x) / (nodes_extended[j + k] - nodes_extended[j + 1])) * basis[:, j + 1]
            basis[:, j] = left + right

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    for j in range(0, n):
        ax.plot(x, basis[:, j])

    plt.show()

def profile(model, device, gradients):
    x = torch.rand(10, 48, 256, 256, device=device, requires_grad=gradients)

    with torch.profiler.profile() as prof:
        y = model(x)
        if gradients:
            dy_dx = torch.autograd.grad(inputs=x, outputs=torch.sum(y))

    with torch.profiler.profile() as prof:
        y = model(x)
        if gradients:
            dy_dx = torch.autograd.grad(inputs=x, outputs=torch.sum(y))

    print(prof.key_averages().table())

if __name__ == '__main__':
    main()
    # b_splines()



import torch

from .ImageFilter import ImageFilter
from .potential import Potential

class FieldsOfExperts(torch.nn.Module):
    """
    Class representing fields of experts (short: FoE) model which is used as regulariser. It consists
    of the components
        > filters
        > potential
    """
    def __init__(self, potential: Potential, image_filter: ImageFilter) -> None:
        """
        Initialisation of an FoE-model.

        :param potential: Object of class Potential
        :param image_filter: Object of class ImageFilter
        """
        super().__init__()

        self.potential = potential
        self.image_filter = image_filter

    def get_image_filter(self) -> ImageFilter:
        return self.image_filter

    def get_potential(self) -> Potential:
        return self.potential

    def forward(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:
        x_conv = self.image_filter(x)
        return self.potential.forward(x_conv, reduce=reduce)




from pylopt.scheduler.HyperParamScheduler import (HyperParamScheduler, CosineAnnealingLRScheduler,
                                                  NAGLipConstGuard, AdaptiveLRRestartScheduler)
from pylopt.scheduler.restart_policy import restart_condition_loss_based, restart_condition_gradient_based


import torch
import numpy as np
import logsumexpv2 as lse
import math

from bilevel_optimisation.utils.Timer import Timer

@torch.compile()
def quartic_sub_function_0(x: torch.Tensor) -> torch.Tensor:
	z = x + 0.5
	return (11 + z * (12 + z * (-6 + z * (-12 + 6 * z)))) / 24

@torch.compile()
def quartic_sub_function_1(x: torch.Tensor) -> torch.Tensor:
	z = 1.5 - x
	return (1 + z * (4 + z * (6 + z * (4 - 4 * z)))) / 24

@torch.compile()
def quartic_sub_function_2(x: torch.Tensor) -> torch.Tensor:
	z = (2.5 - x) * (2.5 - x)
	return (z * z) / 24

@torch.compile()
def quartic_sub_function_3(x: torch.Tensor) -> torch.Tensor:
	return torch.zeros_like(x)

@torch.compile()
def base_quartic_spline_func(t):
	t_abs = torch.abs(t)
	result = quartic_sub_function_0(t_abs)
	result = torch.where(t_abs >= 0.5, quartic_sub_function_1(t_abs), result)
	result = torch.where(t_abs >= 1.5, quartic_sub_function_2(t_abs), result)
	return torch.where(t_abs >= 2.5, quartic_sub_function_3(t_abs), result)

class QuarticSpline(torch.nn.Module):
	def __init__(self, num_marginals, num_nodes, device):
		super().__init__()

		self.num_nodes = num_nodes
		self.num_marginals = num_marginals

		self.box_upper = 3
		self.box_lower = -3

		self.weight_tensor = torch.rand(self.num_marginals, self.num_nodes, device=device)
		self.nodes = torch.nn.Parameter(torch.linspace(self.box_lower, self.box_upper,
                                                       self.num_nodes, device=device).reshape(1, 1, -1, 1, 1),
										requires_grad=False)
		self.scale = (self.box_upper - self.box_lower) / (self.num_nodes - 1)

	def forward(self, x):
		x_scaled = (x.unsqueeze(dim=2) - self.nodes) / self.scale
		# y = torch.einsum('bfnhw, fn->bfhw', torch.exp(x_scaled), torch.exp(self.weight_tensor))
		y = torch.einsum('bfnhw, fn->bfhw', base_quartic_spline_func(x_scaled), torch.exp(self.weight_tensor))
		return torch.sum(torch.log(y + 1e-7))

@torch.compile()
def cubic_sub_function_0(x: torch.Tensor) -> torch.Tensor:
	return 4 - x * (x * (1 + 3 * x))

@torch.compile()
def cubic_sub_function_1(x: torch.Tensor) -> torch.Tensor:
	return (2 - x) * (2 - x) * (2 - x)

@torch.compile()
def cubic_sub_function_2(x: torch.Tensor) -> torch.Tensor:
	return torch.zeros_like(x)




	# result = cubic_sub_function_0(t_abs)
	# result = torch.where(t_abs >= 1, cubic_sub_function_1(t_abs), result)
	# return torch.where(t_abs >= 2, cubic_sub_function_2(t_abs), result)

class CubicSpline(torch.nn.Module):
	def __init__(self, num_marginals, num_nodes, device):
		super().__init__()

		self.num_nodes = num_nodes
		self.num_marginals = num_marginals

		self.box_upper = 3
		self.box_lower = -3

		self.weight_tensor = torch.rand(self.num_marginals, self.num_nodes, device=device)
		self.nodes = torch.nn.Parameter(torch.linspace(self.box_lower, self.box_upper,
													   self.num_nodes, device=device).reshape(1, 1, -1, 1, 1),
										requires_grad=False)
		self.scale = (self.box_upper - self.box_lower) / (self.num_nodes - 1)

	@staticmethod
	def base_cubic_spline_func(t):
		t_abs = torch.abs(t)

		mask_0 = t_abs <= 1
		mask_1 = (1 < t_abs) & (t_abs <= 2)
		mask_2 = t_abs > 2

		return mask_0 * cubic_sub_function_0(t_abs) + mask_1 * cubic_sub_function_1(
			t_abs) + mask_2 * cubic_sub_function_2(t_abs)

		# result = torch.where(
		# 	t_abs <= 1,
		# 	cubic_sub_function_0(t_abs),
		# 	torch.where(
		# 		t_abs <= 2,
		# 		cubic_sub_function_1(t_abs),
		# 		cubic_sub_function_2(t_abs)
		# 	)
		# )

		# return result

	def forward(self, x):
		x_scaled = (x.unsqueeze(dim=2) - self.nodes) / self.scale
		y = torch.einsum('bfnhw, fn->bfhw', CubicSpline.base_cubic_spline_func(x_scaled), torch.exp(self.weight_tensor))
		return torch.log(y + 1e-7) # torch.sum()

class StudentT(torch.nn.Module):
	def __init__(self, num_marginals, device):
		super().__init__()
		self.num_marginals = num_marginals
		self.weight_tensor = torch.rand(self.num_marginals, device=device)

	def forward(self, x):
		return torch.einsum('bfhw,f->', torch.log(1.0 + x ** 2), torch.exp(self.weight_tensor))

class GaussianMixture(torch.nn.Module):

	def __init__(self, num_marginals, num_nodes, device):
		super().__init__()

		self.num_nodes = num_nodes
		self.num_marginals = num_marginals
		self.device = device

		self.box_upper = 3
		self.box_lower = -3

		self.weight_tensor = torch.rand(self.num_marginals, self.num_nodes, device=device)
		self.nodes = torch.nn.Parameter(torch.linspace(self.box_lower, self.box_upper,
													   self.num_nodes, device=device).reshape(1, 1, -1, 1, 1),
										requires_grad=False)
		self.scale = (self.box_upper - self.box_lower) / (self.num_nodes - 1)

	def forward(self, x):
		x_scaled = -0.5 * ((x.unsqueeze(dim=2) - self.nodes) / self.scale) ** 2
		x_weighted = torch.log(self.weight_tensor.unsqueeze(dim=0).unsqueeze(dim=-1).unsqueeze(dim=-1)) + x_scaled
		gaussian_multiplier = 0.5 * torch.log(2 * torch.pi * (self.scale ** 2) * torch.ones(1, device=self.device))
		return torch.sum(-torch.logsumexp(x_weighted - gaussian_multiplier, dim=2))

		# mus = torch.linspace(self.box_lower, self.box_upper, self.num_nodes, device=self.device)
		# mus = mus[None].repeat(48, 1)
		# f_cu, dxf_cu = lse.pot_act(x, self.weight_tensor, mus, self.scale * torch.ones((48, self.num_nodes), device=self.device))
		# return torch.sum(f_cu)



class BumpFunction(torch.nn.Module):
	def __init__(self, num_marginals, num_nodes, device):
		super().__init__()

		self.num_nodes = num_nodes
		self.num_marginals = num_marginals

		self.box_upper = 3
		self.box_lower = -3

		self.weight_tensor = torch.rand(self.num_marginals, self.num_nodes, device=device)
		self.nodes = torch.nn.Parameter(torch.linspace(self.box_lower, self.box_upper,
													   self.num_nodes, device=device).reshape(1, 1, -1, 1, 1),
										requires_grad=False)
		self.scale = (self.box_upper - self.box_lower) / (self.num_nodes - 1)

	@staticmethod
	def _bump_sub_function_0(x):
		return torch.exp(-1.0 / (1 - x ** 2))

	@staticmethod
	def _bump_sub_function_1(x):
		return torch.zeros_like(x)

	@staticmethod
	def _base_bump_func(t):
		t_abs = torch.abs(t)
		result = BumpFunction._bump_sub_function_0(t_abs)
		return torch.where(t_abs >= 1, BumpFunction._bump_sub_function_1(t_abs), result)

	def forward(self, x):
		x_scaled = (x.unsqueeze(dim=2) - self.nodes) / self.scale
		y = torch.einsum('bfnhw, fn->bfhw', BumpFunction._base_bump_func(x_scaled), torch.exp(self.weight_tensor))
		return torch.sum(torch.log(y + 1e-7))

class ClampedExponential(torch.nn.Module):
	def __init__(self, num_marginals, num_nodes, device):
		super().__init__()

		self.num_nodes = num_nodes
		self.num_marginals = num_marginals

		self.box_upper = 3
		self.box_lower = -3

		self.weight_tensor = torch.rand(self.num_marginals, self.num_nodes, device=device)
		self.nodes = torch.nn.Parameter(torch.linspace(self.box_lower, self.box_upper,
													   self.num_nodes, device=device).reshape(1, 1, -1, 1, 1),
										requires_grad=False)
		self.scale = (self.box_upper - self.box_lower) / (self.num_nodes - 1)

		self.support_range = 3 * torch.ones(1, device=device)
		self.min_val = torch.exp(-self.support_range ** 2)

	@torch.compile()
	def _clamped_exp(self, x):
		return torch.clamp(torch.exp(-x**2), min=self.min_val)

	def forward(self, x):
		x_scaled = (x.unsqueeze(dim=2) - self.nodes) / self.scale

		y = torch.einsum('bfnhw, fn->bfhw', self._clamped_exp(x_scaled), torch.exp(self.weight_tensor))
		return torch.sum(torch.log(y + 1e-7))

def profile(model, device, gradients):
	x = torch.rand(10, 48, 256, 256, device=device, requires_grad=gradients)

	with torch.profiler.profile() as prof:
		y = model(x)
		if gradients:
			dy_dx = torch.autograd.grad(inputs=x, outputs=torch.sum(y))

	with torch.profiler.profile() as prof:
		y = model(x)
		if gradients:
			dy_dx = torch.autograd.grad(inputs=x, outputs=torch.sum(y))

	print(prof.key_averages().table())


def measure_performance(model, num_predictions, device, gradients=False):
	x = torch.rand(10, 48, 256, 256, device=device, requires_grad=gradients)

	timings = []
	for _ in range(0, num_predictions):
		with Timer(device) as t:
			y = model(x)
			# if gradients:
			# 	dy_dx = torch.autograd.grad(inputs=x, outputs=y)

		timings.append(t.time_delta())

	print('median: {:.5f}'.format(np.median(timings)))
	print('mean: {:.5f}'.format(np.mean(timings)))
	print('mean (tail): {:.5f}'.format(np.mean(timings[5::])))

def main():

	device = torch.device('cuda')
	num_marginals = 48
	num_nodes = 125
	# spline_quartic = QuarticSpline(num_marginals=num_marginals, num_nodes=num_nodes, device=device)
	# spline_quartic = torch.compile(spline_quartic, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)

	spline_cubic = CubicSpline(num_marginals=num_marginals, num_nodes=num_nodes, device=device)
	spline_cubic = torch.compile(spline_cubic, mode='max-autotune', backend='inductor', dynamic=True,
								 fullgraph=True)
	#
	student_t = StudentT(num_marginals, device)
	student_t = torch.compile(student_t, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)

	# clamped_exp = ClampedExponential(num_marginals=num_marginals, num_nodes=num_nodes, device=device)
	# clamped_exp = torch.compile(clamped_exp, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)
	#
	# bump = BumpFunction(num_marginals=num_marginals, num_nodes=num_nodes, device=device)
	# bump = torch.compile(bump, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)
	#
	gmm = GaussianMixture(num_marginals=num_marginals, num_nodes=num_nodes, device=device)
	gmm = torch.compile(gmm, mode='max-autotune', backend='inductor', dynamic=True, fullgraph=True)
	#
	# x = torch.rand(8, 48, 32, 32, device=device, requires_grad=True)
	# weight_tensor = torch.rand(num_marginals, num_nodes, device=device)
	# mus = torch.linspace(-3, 3, num_nodes, device=device)
	# mus = mus[None].repeat(48, 1)
	# sigma = 2 / (num_nodes - 1)
	# sigmas = sigma * torch.ones((48, num_nodes), device=device)

	# gmm(x)

	print('waaaas')


	# times = []
	# opt = torch.compile(lse.pot_act)
	# for _ in range(0, 100):
	# 	with Timer(device) as t:
	# 		# f_cu, dxf_cu = opt(x, weight_tensor, mus, sigmas)
	# 		gmm(x)
	# 	times.append(t.time_delta())
	#
	# print(np.mean(times))
	# print(np.median(times))





	print('student t (forward + backward)')
	print('------------------------------')
	measure_performance(student_t, 100, device, gradients=True)
	print('')
	#
	# print('quartic spline (forward + backward)')
	# print('---------------------------')
	# measure_performance(spline_quartic, 100, device, gradients=True)

	print('')

	print('cubic spline (forward + backward)')
	print('---------------------------------')
	measure_performance(spline_cubic, 100, device, gradients=True)

	profile(spline_cubic, device, gradients=False)

	print('')

	# print('clamped exponential (forward + backward)')
	# print('---------------------------')
	# measure_performance(clamped_exp, 100, device, gradients=True)
	#
	# print('')
	#
	# print('bump (forward + backward)')
	# print('---------------------------')
	# measure_performance(bump, 100, device, gradients=True)
	#
	# print('')

	print('gmm (forward + backward)')
	print('---------------------------')
	measure_performance(gmm, 100, device, gradients=True)

	print('')
	print('# ### ##############################################################')
	print('')

	# print('student t (forward)')
	# print('-------------------')
	# measure_performance(student_t, 100, device, gradients=False)
	#
	# print('')
	#
	# print('quartic spline (forward)')
	# print('------------------------')
	# measure_performance(spline_quartic, 100, device, gradients=False)
	#
	# print('')
	#
	# print('cubic spline (forward)')
	# print('----------------------')
	# measure_performance(spline_cubic, 100, device, gradients=False)

	# print('')
	#
	# print('clamped exponential (forward)')
	# print('-----------------------------')
	# measure_performance(clamped_exp, 100, device, gradients=False)
	#
	# print('')
	#
	# print('bump (forward)')
	# print('--------------')
	# measure_performance(bump, 100, device, gradients=False)
	#
	# print('')
	#
	# print('gmm (forward)')
	# print('--------------')
	# measure_performance(gmm, 100, device, gradients=False)

if __name__ == '__main__':
	main()



import torch
from torch.utils.data import DataLoader

from bilevel_optimisation.dataset.ImageDataset import ImageDataset
from bilevel_optimisation.utils.DatasetUtils import collate_fn

dtype = torch.float64
train_data_root_dir = '/home/florianthaler/Documents/data/image_data/BSDS300/images/train'
train_image_dataset = ImageDataset(root_path=train_data_root_dir, dtype=dtype)

batch_size = 32
crop_size = 64


train_loader = DataLoader(train_image_dataset, batch_size=batch_size, shuffle=False,
                                  collate_fn=lambda x: collate_fn(x, crop_size=crop_size))

dataset = list(train_loader)
noise_level = 0.1

for i in range(0, 5):
    batch_clean = dataset[i]
    batch_noisy = batch_clean + noise_level * torch.randn_like(batch_clean)

    torch.save(batch_clean, '/home/florianthaler/Documents/research/stochastic_bilevel_optimisation/data/data_batches/batch_clean_{:d}.pt'.format(i))
    torch.save(batch_noisy, '/home/florianthaler/Documents/research/stochastic_bilevel_optimisation/data/data_batches/batch_noisy_{:d}.pt'.format(i))






import torch
from typing import Callable
from typing import Tuple, Dict

from pylopt.data import SolverResult, LinearSolverStats
from pylopt.solver.LinearSystemSolver import IterativeLinearSystemSolver

def scalar_product_l2(x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
    return torch.sum(x1 * x2)

class CGSolver(IterativeLinearSystemSolver):

    def __init__(self, max_num_iterations: int=500, abs_tol: float = 1e-5, scalar_product: Callable = None) -> None:
        super().__init__()

        self._max_num_iterations = max_num_iterations
        self._absolute_tolerance = abs_tol

        self._scalar_product = scalar_product if scalar_product is not None else scalar_product_l2

    def _stopping_criteria_met(self, curr_residual: torch.Tensor,
                              curr_iteration_idx: int, max_num_iterations: int) -> Tuple[bool, Dict]:
        residual_norm = torch.sqrt(self._scalar_product(curr_residual, curr_residual)).cpu().item()
        info = {'max_num_iterations_reached': curr_iteration_idx >= max_num_iterations - 1,
                'residual_accuracy_reached': bool(residual_norm <= self._absolute_tolerance)}
        return info['max_num_iterations_reached'] or info['residual_accuracy_reached'], info

    @torch.no_grad()
    def solve(self, linear_operator: Callable, b: torch.Tensor, x0: torch.Tensor = None) -> SolverResult:
        max_num_iterations = self._max_num_iterations if self._max_num_iterations is not None else b.numel()
        x = torch.zeros_like(b) if x0 is None else x0
        solution = None

        r0 = b - linear_operator(x)
        r = r0.clone()
        p = r.clone()
        residual_norm_list = []
        iterate_list = []
        info = {}
        for k in range(0, max_num_iterations):
            tmp = linear_operator(p)

            alpha = self._scalar_product(r, r) / self._scalar_product(p, tmp)
            x_new = x + alpha * p
            r_new = r - alpha * tmp
            beta = self._scalar_product(r_new, r_new) / self._scalar_product(r, r)
            p_new = r_new + beta * p

            curr_res = torch.sqrt(self._scalar_product(r_new, r_new)).cpu().item()
            residual_norm_list.append(curr_res)
            iterate_list.append(x_new)
            stop, info = self._stopping_criteria_met(r_new, k, max_num_iterations)

            if stop:
                solution = x_new.clone()
                solution = solution.detach()
                break

            x = x_new.clone()
            p = p_new.clone()
            r = r_new.clone()

        cg_stats = LinearSolverStats(num_iterations=len(iterate_list), iterate_list=iterate_list,
                                     residual_norm_list=residual_norm_list)
        return SolverResult(solution=solution, info=info, stats=cg_stats)


import torch
from dataclasses import dataclass
from typing import List, Optional, Dict, Any

@dataclass
class SolverStats:
    num_iterations: int
    iterate_list: List[torch.Tensor]

@dataclass
class LinearSolverStats(SolverStats):
    residual_norm_list: List[float]

@dataclass
class SolverResult:
    solution: Optional[torch.Tensor]
    info: Dict[str, Any]
    stats: SolverStats




import os
import cv2
import shutil

# path_bsd68 = '/home/florianthaler/Documents/trash/denoising_datasets/CBSD68'
# image_path = '/home/florianthaler/Documents/data/image_data/BSDS300/images/test'
# target_path = '/home/florianthaler/Documents/data/image_data/BSDS68'

path_src = '/home/florianthaler/Documents/data/image_data/BSDS68'
path_target = '/home/florianthaler/Documents/data/image_data/BSDS68_rotated'


for item in os.listdir(path_src):

    # base = os.path.splitext(item)[0]
    # shutil.copy2(os.path.join(image_path, '{:s}.jpg'.format(base)), target_path)
    img = cv2.imread(os.path.join(path_src, item))
    h, w, c = img.shape
    if h < w:
        img = cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE)
    cv2.imwrite(os.path.join(path_target, item), img)



import scipy
import torch
import os
import numpy as np

path = '/home/florianthaler/Documents/research/bilevel_optimisation/bilevel_optimisation/model_data/'
data = scipy.io.loadmat(os.path.join(path, 'foe_7x7.mat'))

filters_scaled = torch.load(os.path.join(path, 'foe_filters_7x7_chen-ranftl-pock_2014_scaled.pt'))
thetas_new = torch.load(os.path.join(path, 'student_t_potential_thaler_2025.pt'))


filters = data['filters']
thetas = data['theta'] / 255 ** 2

thetas_new['state_dict']['weight_tensor'] = torch.log(torch.from_numpy(thetas)).squeeze()

torch.save(thetas_new, os.path.join(path, 'foe_thetas_7x7_chen-ranftl-pock_2014.pt'))

print('asd')